#=========================================================================
# VerilogVerilatorImportPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 25, 2019
"""Provide a pass that imports arbitrary SystemVerilog modules."""

import copy
import importlib
import json
import linecache
import os
import shutil
import subprocess
import sys
import timeit
from functools import reduce
from importlib import reload
from itertools import cycle
from textwrap import indent

from pymtl3 import MetadataKey
from pymtl3.datatypes import Bits, is_bitstruct_class, is_bitstruct_inst, mk_bits
from pymtl3.dsl import Component
from pymtl3.dsl.errors import UnsetMetadataError
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.rtype.RTLIRDataType import _get_rtlir_dtype_struct

from ..errors import VerilogImportError
from ..util.utility import (
    expand,
    gen_mapped_ports,
    get_component_unique_name,
    get_rtype,
    make_indent,
    wrap,
)
from ..VerilogPlaceholderPass import VerilogPlaceholderPass
from .verilator_wrapper_c_template import template as c_template
from .verilator_wrapper_py_template import template as py_template


class VerilogVerilatorImportPass( BasePass ):
  """Import an arbitrary SystemVerilog module as a PyMTL component."""

  # Import pass input pass data

  #: Enable import on a component.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  enable              = MetadataKey(bool)

  #: Print out extra debug information during import.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  verbose             = MetadataKey(bool)

  #: Spend more time on compilation for faster simulation performance
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  fast                = MetadataKey(bool)

  #: Use Verilog ``line_trace`` output as the Python line_trace return value.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_line_trace       = MetadataKey(bool)

  #: Enable Verilator coverage.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_coverage         = MetadataKey(bool)

  #: Enable Verilator line coverage.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_line_coverage    = MetadataKey(bool)

  #: Enable Verilator toggle coverage.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_toggle_coverage  = MetadataKey(bool)

  #: Specify the Verilator make directory.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: obj_<top-component-name>
  vl_mk_dir           = MetadataKey(str)

  #: Enable Verilog assert.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_enable_assert    = MetadataKey(bool)

  #: Verilator optimization level.
  #:
  #: Type: ``int``; input
  #:
  #: Default value: ``3``
  vl_opt_level        = MetadataKey(int)

  #: Verilator unroll count.
  #:
  #: Type: ``int``; input
  #:
  #: Default value: ``1000000``
  vl_unroll_count     = MetadataKey(int)

  #: Verilator unroll statement count.
  #:
  #: Type: ``int``; input
  #:
  #: Default value: ``1000000``
  vl_unroll_stmts     = MetadataKey(int)

  #: Enable Verilator lint warnings.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``True``
  vl_W_lint           = MetadataKey(bool)

  #: Enable Verilator style warnings.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``True``
  vl_W_style          = MetadataKey(bool)

  #: Enable Verilator fatal warnings.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``True``
  vl_W_fatal          = MetadataKey(bool)

  #: A list of suppressed Verilator warnings.
  #:
  #: Type: ``[str]``; input
  #:
  #: Default value: ``['UNSIGNED', 'UNOPTFLAT', 'WIDTH']``
  vl_Wno_list         = MetadataKey(list)

  #: Verilator initialization options.
  #:
  #: Possible values: ``'ones'``, ``'zeros'``, ``'rand'``, and non-zero integers; input
  #:
  #: Default value: ``'zeros'``
  vl_xinit            = MetadataKey(str)

  #: Enable Verilator VCD tracing.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_trace            = MetadataKey(bool)

  #: Filename of Verilator VCD tracing.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: translated-component-name.verilator1
  vl_trace_filename   = MetadataKey(str)

  #: Time scale of generated Verilator VCD.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ``'10ps'``
  vl_trace_timescale  = MetadataKey(str)

  #: Cycle time of PyMTL clk pin in the generated Verilator VCD in unit of ``vl_trace_timescale``.
  #:
  #: Type: ``int``; input
  #:
  #: Default value: ``100``
  vl_trace_cycle_time = MetadataKey(int)

  #: Set to true to allow for on-demand VCD dumping
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  vl_trace_on_demand  = MetadataKey(bool)

  #: Top level port name that is used to enable VCD dumping when
  #: `vl_trace_on_demand` is True. Assuming the port is an active-high
  #: enable signal.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ``""``
  vl_trace_on_demand_portname = MetadataKey(str)

  #: Optional flags to be passed to the C compiler.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ``''``
  c_flags             = MetadataKey(str)

  #: Optional include paths to be passed to the C compiler.
  #:
  #: Type: ``[str]``; input
  #:
  #: Default value: ``[]``
  c_include_path      = MetadataKey(list)

  #: Optional source file paths to be passed to the C compiler.
  #:
  #: Type: ``[str]``; input
  #:
  #: Default value: ``[]``
  c_srcs              = MetadataKey(list)

  #: Optional flags to be passed to LD.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ``''``
  ld_flags            = MetadataKey(str)

  #: Optional libraries to be passed to LD (e.g., ``'-lfoo'``).
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ``''``
  ld_libs             = MetadataKey(str)

  # Import pass output pass data

  #: An instnace of :class:`VerilatorImportConfigs` containing the parsed options.
  #:
  #: Type: :class:`VerilatorImportConfigs`; output
  import_config       = MetadataKey()

  def __call__( s, top ):
    """Import the PyMTL component hierarhcy rooted at ``top``."""
    s.top = top
    if not top._dsl.constructed:
      raise VerilogImportError( top,
        f"please elaborate design {top} before applying the import pass!" )
    ret = s.traverse_hierarchy( top )
    if ret is None:
      ret = top
    else:
      ret.elaborate()
    return ret

  def traverse_hierarchy( s, m ):
    c = s.__class__
    ph_pass = c.get_placeholder_pass()
    # Import can only be performed on Placeholders
    if m.has_metadata( ph_pass.enable ) and m.get_metadata( ph_pass.enable ):
      m.set_metadata( c.import_config, c.get_import_config()( m ) )
      return s.do_import( m )

    else:
      for child in m.get_child_components(repr):
        s.traverse_hierarchy( child )

  def do_import( s, m ):
    try:
      imp = s.get_imported_object( m )
      if m is s.top:
        return imp
      else:
        s.top.replace_component_with_obj( m, imp )
    except AssertionError as e:
      msg = '' if e.args[0] is None else e.args[0]
      raise VerilogImportError( m, msg )

  #-----------------------------------------------------------------------
  # Backend-specific methods
  #-----------------------------------------------------------------------

  @staticmethod
  def get_backend_name():
    return "verilog"

  @staticmethod
  def get_placeholder_pass():
    return VerilogPlaceholderPass

  @staticmethod
  def get_translation_pass():
    from ..translation.VerilogTranslationPass import VerilogTranslationPass
    return VerilogTranslationPass

  @staticmethod
  def get_import_config():
    from .VerilogVerilatorImportConfigs import VerilogVerilatorImportConfigs
    return VerilogVerilatorImportConfigs

  @staticmethod
  def get_gen_mapped_port():
    return gen_mapped_ports

  #-----------------------------------------------------------------------
  # is_cached
  #-----------------------------------------------------------------------

  def is_cached( s, m, ip_cfg ):
    c = s.__class__
    # Only components translated by pymtl translation passes will be cached
    try:
      is_same = m.get_metadata( c.get_translation_pass().is_same )
    except UnsetMetadataError:
      is_same = False

    # Check if the verilated model is cached
    is_source_cached = False
    obj_dir = ip_cfg.vl_mk_dir
    c_wrapper = ip_cfg.get_c_wrapper_path()
    py_wrapper = ip_cfg.get_py_wrapper_path()
    shared_lib = ip_cfg.get_shared_lib_path()
    if is_same and os.path.exists(obj_dir) and os.path.exists(c_wrapper) and \
       os.path.exists(py_wrapper) and os.path.exists(shared_lib):
      is_source_cached = True

    # Check if the configurations from the last run are the same
    is_config_cached = False
    config_file = f'pymtl_import_config_{ip_cfg.translated_top_module}.json'
    new_cfg = s.serialize_cfg( ip_cfg )
    if os.path.exists(config_file):
      with open(config_file) as fd:
        if s.is_same_cfg( json.load( fd ), new_cfg ):
          is_config_cached = True

    return is_source_cached and is_config_cached, config_file, new_cfg

  #-----------------------------------------------------------------------
  # get_imported_object
  #-----------------------------------------------------------------------

  def get_imported_object( s, m ):
    c = s.__class__
    ph_cfg = m.get_metadata( c.get_placeholder_pass().placeholder_config )
    ip_cfg = m.get_metadata( c.import_config )
    ip_cfg.setup_configs( m, c.get_translation_pass(), c.get_placeholder_pass() )

    rtype = RTLIRGetter(cache=False).get_component_ifc_rtlir( m )

    # Now we selectively unpack array of ports if they are referred to in
    # port_map
    ports = c.get_gen_mapped_port()( m, ph_cfg.port_map,
                                     ph_cfg.has_clk, ph_cfg.has_reset,
                                     ph_cfg.separator )

    cached, config_file, cfg_d = s.is_cached( m, ip_cfg )

    s.create_verilator_model( m, ph_cfg, ip_cfg, cached )

    port_cdefs = s.create_verilator_c_wrapper( m, ph_cfg, ip_cfg, ports, cached )

    s.create_shared_lib( m, ph_cfg, ip_cfg, cached )

    symbols = s.create_py_wrapper( m, ph_cfg, ip_cfg, rtype, ports, port_cdefs, cached )

    imp = s.import_component( m, ph_cfg, ip_cfg, symbols )

    imp._ip_cfg = ip_cfg
    imp._ph_cfg = ph_cfg
    imp._ports = ports

    # Dump configuration dict to config_file
    with open( config_file, 'w' ) as fd:
      json.dump( cfg_d, fd, indent = 4 )

    return imp

  #-----------------------------------------------------------------------
  # create_verilator_model
  #-----------------------------------------------------------------------

  def create_verilator_model( s, m, ph_cfg, ip_cfg, cached ):
    # Verilate module `m`.

    ip_cfg.vprint("\n=====Verilate model=====")

    if not cached:
      # Generate verilator command
      cmd = ip_cfg.create_vl_cmd()

      # Remove obj_dir directory if it already exists.
      # obj_dir is where the verilator output ( C headers and sources ) is stored
      obj_dir = ip_cfg.vl_mk_dir
      if os.path.exists( obj_dir ):
        shutil.rmtree( obj_dir )

      succeeds = True

      # Try to call verilator
      try:
        ip_cfg.vprint(f"Verilating {ip_cfg.translated_top_module} with command:", 2)
        ip_cfg.vprint(f"{cmd}", 4)
        t0 = timeit.default_timer()
        subprocess.check_output(
            cmd, stderr = subprocess.STDOUT, shell = True )
        ip_cfg.vprint(f"verilate time: {timeit.default_timer()-t0}")
      except subprocess.CalledProcessError as e:
        succeeds = False
        err_msg = e.output if not isinstance(e.output, bytes) else \
                  e.output.decode('utf-8')
        import_err_msg = \
            f"Fail to verilate model {ip_cfg.translated_top_module}\n"\
            f"  Verilator command:\n{indent(cmd, '  ')}\n\n"\
            f"  Verilator output:\n{indent(wrap(err_msg), '  ')}\n"

      if not succeeds:
        raise VerilogImportError(m, import_err_msg)

      ip_cfg.vprint("Successfully verilated the given model!", 2)

    else:
      ip_cfg.vprint(f"{ip_cfg.translated_top_module} not verilated because it's cached!", 2)

  #-----------------------------------------------------------------------
  # create_verilator_c_wrapper
  #-----------------------------------------------------------------------

  def create_verilator_c_wrapper( s, m, ph_cfg, ip_cfg, ports, cached ):
    # Return the file name of generated C component wrapper.
    # Create a C wrapper that calls verilator C API and provides interfaces
    # that can be later called through CFFI.

    component_name = ip_cfg.translated_top_module
    vl_component_name = s._verilator_name(component_name)
    dump_vcd = int(ip_cfg.vl_trace)
    vcd_timescale = ip_cfg.vl_trace_timescale
    half_cycle_time = ip_cfg.vl_trace_cycle_time // 2
    external_trace = int(ip_cfg.vl_line_trace)
    wrapper_name = ip_cfg.get_c_wrapper_path()
    verilator_xinit_value = ip_cfg.get_vl_xinit_value()
    verilator_xinit_seed = ip_cfg.get_vl_xinit_seed()
    has_clk = int(ph_cfg.has_clk)

    # On-demand VCD dumping configs
    on_demand_dump_vcd = int(ip_cfg.vl_trace_on_demand)
    on_demand_vcd_enable = str(ip_cfg.vl_trace_on_demand_portname)
    if on_demand_vcd_enable:
      on_demand_vcd_enable = f"model->{on_demand_vcd_enable}"
    else:
      on_demand_vcd_enable = 0

    ip_cfg.vprint("\n=====Generate C wrapper=====")

    # Generate port declarations for the verilated model in C
    port_defs = []
    for _, v_name, port, _ in ports:
      if v_name:
        port_defs.append( s.gen_signal_decl_c( v_name, port ) )
    port_cdefs = copy.copy( port_defs )
    make_indent( port_defs, 2 )
    port_defs = '\n'.join( port_defs )

    # Generate initialization statements for in/out ports
    port_inits = []
    for _, v_name, port, _ in ports:
      if v_name:
        port_inits.extend( s.gen_signal_init_c( v_name, port ) )
    make_indent( port_inits, 1 )
    port_inits = '\n'.join( port_inits )

    # Fill in the C wrapper template
    with open( wrapper_name, 'w' ) as output:
      output.write( c_template.format( **locals() ) )

    ip_cfg.vprint(f"Successfully generated C wrapper {wrapper_name}!", 2)
    return port_cdefs

  #-----------------------------------------------------------------------
  # create_shared_lib
  #-----------------------------------------------------------------------

  def create_shared_lib( s, m, ph_cfg, ip_cfg, cached ):
    # Return the name of compiled shared lib.

    full_name = ip_cfg.translated_top_module
    dump_vcd = ip_cfg.vl_trace
    ip_cfg.vprint("\n=====Compile shared library=====")

    if not cached:
      cmd = ip_cfg.create_cc_cmd()

      succeeds = True

      # Try to call the C compiler
      try:
        ip_cfg.vprint("Compiling shared library with command:", 2)
        ip_cfg.vprint(f"{cmd}", 4)
        t0 = timeit.default_timer()
        subprocess.check_output(
            cmd,
            stderr = subprocess.STDOUT,
            shell = True,
            universal_newlines = True
        )
        ip_cfg.vprint(f"shared library compilation time: {timeit.default_timer()-t0}")
      except subprocess.CalledProcessError as e:
        succeeds = False
        err_msg = e.output if not isinstance(e.output, bytes) else \
                  e.output.decode('utf-8')
        import_err_msg = \
            f"Failed to compile Verilated model into a shared library:\n"\
            f"  C compiler command:\n{indent(cmd, '  ')}\n\n"\
            f"  C compiler output:\n{indent(wrap(err_msg), '  ')}\n"

      if not succeeds:
        raise VerilogImportError(m, import_err_msg)

      ip_cfg.vprint(f"Successfully compiled shared library "\
                    f"{ip_cfg.get_shared_lib_path()}!", 2)

    else:
      ip_cfg.vprint("Didn't compile shared library because it's cached!", 2)

  #-----------------------------------------------------------------------
  # create_py_wrapper
  #-----------------------------------------------------------------------
  # Return the file name of the generated PyMTL component wrapper.

  def create_py_wrapper( s, m, ph_cfg, ip_cfg, rtype, ports, port_cdefs, cached ):
    ip_cfg.vprint("\n=====Generate PyMTL wrapper=====")

    wrapper_name = ip_cfg.get_py_wrapper_path()

    # Port definitions of verilated model
    make_indent( port_cdefs, 4 )

    # Port definition in PyMTL style
    symbols, port_defs = s.gen_signal_decl_py( rtype )
    make_indent( port_defs, 2 )

    # Set upblk inputs and outputs
    set_comb_input, structs_input   = s.gen_comb_input( ports, symbols )
    set_comb_output, structs_output = s.gen_comb_output( ports, symbols )
    make_indent( structs_input, 2 )
    make_indent( structs_output, 2 )
    make_indent( set_comb_input, 3 )
    make_indent( set_comb_output, 3 )

    # Line trace
    line_trace = s.gen_line_trace_py( ports )

    # Internal line trace
    in_line_trace = s.gen_internal_line_trace_py( ports )

    # External trace function definition
    if ip_cfg.vl_line_trace:
      external_trace_c_def = f'void trace( V{ip_cfg.translated_top_module}_t *, char * );'
    else:
      external_trace_c_def = ''

    # Fill in the python wrapper template
    with open( wrapper_name, 'w' ) as output:
      py_wrapper = py_template.format(
        component_name        = ip_cfg.translated_top_module,
        has_clk               = int(ph_cfg.has_clk),
        clk                   = 'inv_clk' if not ph_cfg.has_clk else \
                                next(filter(lambda x: x[0][0]=='clk', ports))[1],
        lib_file              = ip_cfg.get_shared_lib_path(),
        port_cdefs            = ('  '*4+'\n').join( port_cdefs ),
        port_defs             = '\n'.join( port_defs ),
        structs_input         = '\n'.join( structs_input ),
        structs_output        = '\n'.join( structs_output ),
        set_comb_input        = '\n'.join( set_comb_input ),
        set_comb_output       = '\n'.join( set_comb_output ),
        line_trace            = line_trace,
        in_line_trace         = in_line_trace,
        dump_vcd              = int(ip_cfg.vl_trace),
        has_vl_trace_filename = bool(ip_cfg.vl_trace_filename),
        vl_trace_filename     = ip_cfg.vl_trace_filename,
        external_trace        = int(ip_cfg.vl_line_trace),
        trace_c_def           = external_trace_c_def,
      )
      output.write( py_wrapper )

    ip_cfg.vprint(f"Successfully generated PyMTL wrapper {wrapper_name}!", 2)
    return symbols

  #-----------------------------------------------------------------------
  # import_component
  #-----------------------------------------------------------------------
  # Return the PyMTL component imported from `wrapper_name`.v.

  def import_component( s, m, ph_cfg, ip_cfg, symbols ):
    ip_cfg.vprint("=====Create python object=====")

    component_name = ip_cfg.translated_top_module
    # Get the name of the wrapper Python module
    wrapper_name = ip_cfg.get_py_wrapper_path()
    wrapper = wrapper_name.split('.')[0]

    # Add CWD to sys.path so we can import from the current directory
    if not os.getcwd() in sys.path:
      sys.path.append( os.getcwd() )

    # Check linecache in case the wrapper file has been modified
    linecache.checkcache()

    # Import the component from python wrapper

    if wrapper in sys.modules:
      # Reload the wrapper module in case the user has updated the wrapper
      reload(sys.modules[wrapper])
    else:
      # importlib.import_module inserts the wrapper module into sys.modules
      importlib.import_module(wrapper)

    # Try to access the top component class from the wrapper module
    try:
      imp_class = getattr( sys.modules[wrapper], component_name )
    except AttributeError as e:
      raise VerilogImportError(m,
          f"internal error: PyMTL wrapper {wrapper_name} does not have "
          f"top component {component_name}!") from e

    imp = imp_class()
    ip_cfg.vprint(f"Successfully created python object of {component_name}!", 2)

    # Update the global namespace of `construct` so that the struct and interface
    # classes defined previously can still be used in the imported model.
    imp.construct.__globals__.update( symbols )

    ip_cfg.vprint("Import succeeds!")
    return imp

  #-------------------------------------------------------------------------
  # Serialize and compare configurations
  #-------------------------------------------------------------------------

  def serialize_cfg( s, ip_cfg ):
    d = {}
    s._volatile_configs = [
      'verilog_hash',
      'vl_line_trace', 'vl_coverage', 'vl_line_coverage', 'vl_toggle_coverage',
      'vl_mk_dir', 'vl_enable_assert',
      'vl_W_lint', 'vl_W_style', 'vl_W_fatal', 'vl_Wno_list',
      'vl_xinit', 'vl_trace',
      'vl_trace_timescale', 'vl_trace_cycle_time',
      'vl_trace_on_demand', 'vl_trace_on_demand_portname',
      'c_flags', 'c_include_path', 'c_srcs',
      'ld_flags', 'ld_libs',
    ]
    d['ImportPassName'] = 'VerilogVerilatorImportPass'
    for cfg in s._volatile_configs:
      d[cfg] = copy.deepcopy(getattr( ip_cfg, cfg ))

    return d

  def is_same_cfg( s, prev, new ):
    _volatile_configs = copy.copy(s._volatile_configs)
    _volatile_configs.append('ImportPassName')
    if not all(cfg in prev and cfg in new for cfg in _volatile_configs):
      return False
    return all(prev[cfg] == new[cfg] for cfg in _volatile_configs)

  #-------------------------------------------------------------------------
  # gen_signal_decl_c
  #-------------------------------------------------------------------------
  # Return C variable declaration of `port`.

  def gen_signal_decl_c( s, name, port ):

    c_dim = s._get_c_dim( port )
    nbits = s._get_c_nbits( port )
    UNSIGNED_8  = 'unsigned char'
    UNSIGNED_16 = 'unsigned short'
    UNSIGNED_32 = 'unsigned int'
    if sys.maxsize > 2**32:
      UNSIGNED_64 = 'unsigned long'
    else:
      UNSIGNED_64 = 'unsigned long long'
    if    nbits <= 8:  data_type = UNSIGNED_8
    elif  nbits <= 16: data_type = UNSIGNED_16
    elif  nbits <= 32: data_type = UNSIGNED_32
    elif  nbits <= 64: data_type = UNSIGNED_64
    else:              data_type = UNSIGNED_32
    name = s._verilator_name( name )
    return f'{data_type} * {name}{c_dim};'

  #-------------------------------------------------------------------------
  # gen_signal_init_c
  #-------------------------------------------------------------------------
  # Return C port variable initialization.

  def gen_signal_init_c( s, name, port ):

    ret       = []
    c_dim     = s._get_c_dim( port )
    nbits     = s._get_c_nbits( port )
    deference = '&' if nbits <= 64 else ''
    name      = s._verilator_name( name )

    if c_dim:
      n_dim_size = s._get_c_n_dim( port )
      sub = ""
      for_template = \
"""\
for ( int i_{idx} = 0; i_{idx} < {dim_size}; i_{idx}++ )
"""
      assign_template = \
"""\
m->{name}{sub} = {deference}model->{name}{sub};
"""

      for idx, dim_size in enumerate( n_dim_size ):
        ret.append( for_template.format( **locals() ) )
        sub += f"[i_{idx}]"

      ret.append( assign_template.format( **locals() ) )

      # Indent the for loop
      for start, dim_size in enumerate( n_dim_size ):
        for idx in range( start + 1, len( n_dim_size ) + 1 ):
          ret[ idx ] = "  " + ret[ idx ]

    else:
      ret.append(f'm->{name} = {deference}model->{name};')

    return ret

  #-------------------------------------------------------------------------
  # gen_signal_decl_py
  #-------------------------------------------------------------------------
  # Return the PyMTL definition of all interface ports of `rtype`.

  def gen_signal_decl_py( s, rtype ):

    #-----------------------------------------------------------------------
    # Methods that generate signal declarations
    #-----------------------------------------------------------------------

    def gen_dtype_str( symbols, dtype ):
      if isinstance( dtype, rdt.Vector ):
        nbits = dtype.get_length()
        Bits_name = f"Bits{nbits}"
        if Bits_name not in symbols and nbits >= 256:
          Bits_class = mk_bits( nbits )
          symbols.update( { Bits_name : Bits_class } )
        return f'Bits{dtype.get_length()}'
      elif isinstance( dtype, rdt.Struct ):
        # It is possible to reuse the existing struct class because its __init__
        # can be called without arguments.
        name, cls = dtype.get_name(), dtype.get_class()
        if name not in symbols:
          symbols.update( { name : cls } )
        return name
      else:
        assert False, f"unrecognized data type {dtype}!"

    def gen_port_decl_py( ports ):
      symbols, decls = {}, []
      for id_, _port in ports:
        if id_ not in ['clk', 'reset']:
          if isinstance( _port, rt.Array ):
            n_dim = _port.get_dim_sizes()
            rhs = "{direction}( {dtype} )"
            port = _port.get_sub_type()
            _n_dim = copy.copy( n_dim )
            _n_dim.reverse()
            for length in _n_dim:
              rhs = f"[ {rhs} for _ in range({length}) ]"
          else:
            rhs = "{direction}( {dtype} )"
            port = _port
          direction = s._get_direction( port )
          dtype = gen_dtype_str( symbols, port.get_dtype() )
          rhs = rhs.format( **locals() )
          decls.append(f"s.{id_} = {rhs}")
      return symbols, decls

    def gen_ifc_decl_py( ifcs ):

      def gen_ifc_str( symbols, ifc ):

        def _get_arg_str( name, obj ):
          # Support common python types and Bits/BitStruct

          if isinstance( obj, (int, bool, str) ):
            return str(obj)

          elif obj is None:
            return 'None'

          elif isinstance( obj, type ) and issubclass( obj, Bits ):
            nbits = obj.nbits
            Bits_name = f"Bits{nbits}"
            if Bits_name not in symbols and nbits >= 256:
              Bits_class = mk_bits( nbits )
              symbols.update( { Bits_name : Bits_class } )
            return Bits_name

          elif isinstance( obj, Bits ):
            nbits = obj.nbits
            value = int(obj)
            Bits_name = f"Bits{nbits}"
            Bits_arg_str = f"{Bits_name}( {value} )"
            if Bits_name not in symbols and nbits >= 256:
              Bits_class = mk_bits( nbits )
              symbols.update( { Bits_name : Bits_class } )
            return Bits_arg_str

          elif is_bitstruct_class(obj):
            # Shunning: get the unique name of bitstruct class
            BitStruct_name = _get_rtlir_dtype_struct( obj() ).get_name()

            if BitStruct_name in symbols:
              assert symbols[BitStruct_name] is obj, "Same bitstruct name but different classes!"
            else:
              symbols.update( { BitStruct_name : obj } )

            return BitStruct_name

          # FIXME formalize this
          elif isinstance( obj, type ) and hasattr( obj, 'req' ) and hasattr( obj, 'resp' ):
            symbols.update( { obj.__name__ : obj } )
            return obj.__name__

          elif is_bitstruct_inst( obj ):
            raise TypeError("Do you really want to pass in an instance of "
                            "a BitStruct? Contact PyMTL developers!")
            # This is hacky: we don't know how to construct an object that
            # is the same as `obj`, but we do have the object itself. If we
            # add `obj` to the namespace of `construct` everything works fine
            # but the user cannot tell what object is passed to the constructor
            # just from the code.
            # Do not use a double underscore prefix because that will be
            # interpreted differently by the Python interpreter
            # bs_name = ("_" if name[0] != "_" else "") + name + "_obj"
            # if bs_name not in symbols:
              # symbols.update( { bs_name : obj } )
            # return bs_name

          raise TypeError( f"Interface constructor argument {obj} is not an int/Bits/BitStruct/TypeTuple!" )

        name, cls = ifc.get_name(), ifc.get_class()
        if name not in symbols:
          symbols.update( { name : cls } )
        arg_list = []
        args = ifc.get_args()
        for idx, obj in enumerate(args[0]):
          arg_list.append( _get_arg_str( f"_ifc_arg{idx}", obj ) )
        for arg_name, arg_obj in args[1].items():
          arg_list.append( f"{arg_name} = {_get_arg_str( arg_name, arg_obj )}" )
        return name, ', '.join( arg_list )

      symbols, decls = {}, []
      for id_, ifc in ifcs:
        if isinstance( ifc, rt.Array ):
          n_dim = ifc.get_dim_sizes()
          rhs = "{ifc_class}({ifc_params})"
          _ifc = ifc.get_sub_type()
          _n_dim = copy.copy( n_dim )
          _n_dim.reverse()
          for length in _n_dim:
            rhs = f"[ {rhs} for _ in range({length}) ]"
        else:
          rhs = "{ifc_class}({ifc_params})"
          _ifc = ifc
        ifc_class, ifc_params = gen_ifc_str( symbols, _ifc )
        if ifc_params:
          ifc_params = " " + ifc_params + " "
        rhs = rhs.format( **locals() )
        decls.append(f"s.{id_} = {rhs}")
      return symbols, decls

    #-----------------------------------------------------------------------
    # Method gen_signal_decl_py
    #-----------------------------------------------------------------------

    ports = rtype.get_ports_packed()
    ifcs = rtype.get_ifc_views_packed()

    p_symbols, p_decls = gen_port_decl_py( ports )
    i_symbols, i_decls = gen_ifc_decl_py( ifcs )

    return {**p_symbols, **i_symbols}, p_decls + i_decls

  #-----------------------------------------------------------------------
  # Methods that generate python signal writes
  #-----------------------------------------------------------------------

  def _gen_vector_write( s, d, lhs, rhs, dtype, pos ):
    nbits = dtype.get_length()
    l, r = pos, pos+nbits
    if d == 'i':
      ret = [ f"{rhs}[{l}:{r}] @= {lhs}" ]
    else:
      ret = [ f"{lhs} @= {rhs}[{l}:{r}]" ]
    return ret, r

  def _gen_struct_write( s, d, lhs, rhs, dtype, pos ):
    ret = []
    all_properties = reversed(list(dtype.get_all_properties().items()))
    for name, field in all_properties:
      _ret, pos = s._gen_write_dispatch( d, f"{lhs}.{name}", rhs, field, pos )
      ret.extend( _ret )
    return ret, pos

  def _gen_packed_array_write( s, d, lhs, rhs, dtype, n_dim, pos ):
    if not n_dim:
      return s._gen_write_dispatch( d, lhs, rhs, dtype, pos )
    # Recursively generate array
    ret = []
    for idx in range(n_dim[0]):
      _ret, pos = s._gen_packed_array_write( d, f"{lhs}[{idx}]", rhs, dtype, n_dim[1:], pos )
      ret.extend( _ret )
    return ret, pos

  def _gen_write_dispatch( s, d, lhs, rhs, dtype, pos ):
    if isinstance( dtype, rdt.Vector ):
      return s._gen_vector_write( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.Struct ):
      return s._gen_struct_write( d, lhs, rhs, dtype, pos )
    elif isinstance( dtype, rdt.PackedArray ):
      n_dim = dtype.get_dim_sizes()
      sub_dtype = dtype.get_sub_dtype()
      return s._gen_packed_array_write( d, lhs, rhs, sub_dtype, n_dim, pos )
    assert False, f"unrecognized data type {dtype}!"

  #-------------------------------------------------------------------------
  # gen_comb_input
  #-------------------------------------------------------------------------

  def gen_port_vector_input( s, lhs, rhs, mangled_rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    blocks   = [ '',
                 f's.{mangled_rhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@update',
                 f'def isignal_{mangled_rhs}():',
                 f'  s.{mangled_rhs} @= {rhs}' ]
    set_comb = ( s._gen_ref_write( lhs, 's.'+mangled_rhs, dtype_nbits ) )
    return set_comb, blocks

  def gen_port_struct_input( s, lhs, rhs, mangled_rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    # If the top-level signal is a struct, we add the datatype to symbol?
    dtype_name = dtype.get_class().__name__
    if dtype_name not in symbols:
      symbols[dtype_name] = dtype.get_class()

    blocks   = [ '',
                 f's.{mangled_rhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@update',
                 f'def istruct_{mangled_rhs}():',
                 f'  s.{mangled_rhs} @= {rhs}' ]

    # We don't create a new struct if we are copying values from pymtl
    # land to verilator, i.e. this port is the input to the imported
    # component.
    # At the end, we write tmp to the corresponding CFFI variable
    set_comb = s._gen_ref_write( lhs, 's.'+mangled_rhs, dtype_nbits )
    return set_comb, blocks

  def gen_port_input( s, lhs, rhs, pnames, dtype, symbols ):
    rhs = rhs.format(next(pnames))

    # We always name mangle now
    mangled_rhs = s._pymtl_name_mangle( rhs )

    if isinstance( dtype, rdt.Vector ):
      return s.gen_port_vector_input( lhs, rhs, mangled_rhs, dtype, symbols )

    elif isinstance( dtype, rdt.Struct ):
      return s.gen_port_struct_input( lhs, rhs, mangled_rhs, dtype, symbols )

    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_port_array_input( s, lhs, rhs, pnames, dtype, index, n_dim, symbols ):
    if not n_dim:
      return s.gen_port_input( lhs, rhs, pnames, dtype, symbols )
    else:
      set_comb, structs = [], []
      for idx in range( n_dim[0] ):
        _lhs = f"{lhs}[{idx}]"
        if index == 0:
          _rhs = f"{rhs}[{idx}]"
          _index = index
        else:
          _rhs = f"{rhs}"
          _index = index-1
        _set_comb, _structs = s.gen_port_array_input( _lhs, _rhs, pnames, dtype, _index, n_dim[1:], symbols )
        set_comb += _set_comb
        structs  += _structs
      return set_comb, structs

  def gen_comb_input( s, packed_ports, symbols ):
    set_comb, structs = [], []
    # Read all input ports ( except for 'clk' ) from component ports into
    # the verilated model. We do NOT want `clk` signal to be read into
    # the verilated model because only the sequential update block of
    # the imported component should manipulate it.

    for _pnames, vname, rtype, port_idx in packed_ports:
      if isinstance( rtype, rt.Array ):
        n_dim = rtype.get_dim_sizes()
      else:
        n_dim = []
      repeats = reduce(lambda a, b: a*b, n_dim[port_idx:], 1)
      pnames = [name for name in _pnames for _ in range(repeats)]
      pnames_iter = cycle(pnames)
      p_n_dim, p_rtype = get_rtype( rtype )
      if s._get_direction( p_rtype ) == 'InPort' and pnames[0] != 'clk' and vname:
        dtype = p_rtype.get_dtype()
        lhs = "_ffi_m."+s._verilator_name(vname)
        rhs = "s.{}"
        idx = port_idx
        _set_comb, _structs = s.gen_port_array_input( lhs, rhs, pnames_iter, dtype, idx, p_n_dim, symbols )
        set_comb += _set_comb
        structs  += _structs

    return set_comb, structs

  #-------------------------------------------------------------------------
  # gen_comb_output
  #-------------------------------------------------------------------------

  def gen_port_vector_output( s, lhs, mangled_lhs, rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    blocks   = [ '',
                 f's.{mangled_lhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@update',
                 f'def osignal_{mangled_lhs}():',
                 f'  {lhs} @= s.{mangled_lhs}' ]

    set_comb = s._gen_ref_read( 's.'+mangled_lhs, rhs, dtype_nbits, '@=' )
    return set_comb, blocks

  def gen_port_struct_output( s, lhs, mangled_lhs, rhs, dtype, symbols ):
    dtype_nbits = dtype.get_length()
    # If the top-level signal is a struct, we add the datatype to symbol?
    dtype_name = dtype.get_name()
    if dtype_name not in symbols:
      symbols[dtype_name] = dtype.get_class()

    blocks   = [ '',
                 f's.{mangled_lhs} = Wire( {s._gen_bits_decl(dtype_nbits)} )',
                 '@update',
                 f'def ostruct_{mangled_lhs}():' ]

    # We create a long Bits object to accept CFFI value for struct
    # the temporary wire name

    # We create a new struct if we are copying values from verilator
    # world to pymtl land and send it out through the output of this
    # component
    body, pos = s._gen_struct_write( 'o', lhs, 's.'+mangled_lhs, dtype, 0 )
    assert pos == dtype.get_length()

    upblk_content = body
    make_indent( upblk_content, 1 )

    blocks += upblk_content

    # We create a long Bits object tmp first
    # Then we load the full Bits to tmp
    set_comb = s._gen_ref_read( 's.'+mangled_lhs, rhs, dtype_nbits, '@=' )
    return set_comb, blocks

  def gen_port_output( s, lhs, pnames, rhs, dtype, symbols ):
    lhs = lhs.format(next(pnames))

    mangled_lhs = s._pymtl_name_mangle( lhs )

    if isinstance( dtype, rdt.Vector ):
      return s.gen_port_vector_output( lhs, mangled_lhs, rhs, dtype, symbols )
    elif isinstance( dtype, rdt.Struct ):
      return s.gen_port_struct_output( lhs, mangled_lhs, rhs, dtype, symbols )
    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_port_array_output( s, lhs, pnames, rhs, dtype, index, n_dim, symbols ):
    if not n_dim:
      return s.gen_port_output( lhs, pnames, rhs, dtype, symbols )
    else:
      set_comb, structs = [], []
      for idx in range( n_dim[0] ):
        if index == 0:
          _lhs = f"{lhs}[{idx}]"
          _index = index
        else:
          _lhs = f"{lhs}"
          _index = index-1
        _rhs = f"{rhs}[{idx}]"
        _set_comb, _structs = s.gen_port_array_output( _lhs, pnames, _rhs, dtype, _index, n_dim[1:], symbols )
        set_comb += _set_comb
        structs  += _structs
      return set_comb, structs

  def gen_comb_output( s, packed_ports, symbols ):
    set_comb, structs = [], []
    for _pnames, vname, rtype, port_idx in packed_ports:
      if isinstance( rtype, rt.Array ):
        n_dim = rtype.get_dim_sizes()
      else:
        n_dim = []
      repeats = reduce(lambda a, b: a*b, n_dim[port_idx:], 1)
      pnames = [name for name in _pnames for _ in range(repeats)]
      pnames_iter = cycle(pnames)
      p_n_dim, p_rtype = get_rtype( rtype )
      if s._get_direction( rtype ) == 'OutPort':
        dtype = p_rtype.get_dtype()
        lhs = "s.{}"
        rhs = "_ffi_m." + s._verilator_name(vname)
        idx = port_idx
        _set_comb, _structs = s.gen_port_array_output( lhs, pnames_iter, rhs, dtype, idx, p_n_dim, symbols )
        set_comb += _set_comb
        structs  += _structs
    return set_comb, structs

  #-------------------------------------------------------------------------
  # gen_line_trace_py
  #-------------------------------------------------------------------------
  # Return the line trace method body that shows all interface ports.

  def gen_line_trace_py( s, packed_ports ):
    template = '{0}={{s.{0}}},'
    trace_string = ''
    for pnames, _, _, _ in packed_ports:
      for pname in pnames:
        trace_string += ' ' + template.format( pname )
    return f"      return f'{trace_string}'"

  #-------------------------------------------------------------------------
  # gen_internal_line_trace_py
  #-------------------------------------------------------------------------
  # Return the line trace method body that shows all CFFI ports.
  # Now that there could be multiple pnames that correspond to one vname,
  # I'm not sure how to generate internal line trace... maybe we should
  # deprecate internal_line_trace since it's not used by many anyways?

  def gen_internal_line_trace_py( s, packed_ports ):
    # ret = [ '_ffi_m = s._ffi_m', 'lt = ""' ]
    # template = \
    #   "lt += '{vname} = {{}}, '.format(full_vector(s.{pname}, _ffi_m.{vname}))"
    # for pname, vname, port in packed_ports:
    #   if vname:
    #     pname = s._verilator_name(pname)
    #     vname = s._verilator_name(vname)
    #     ret.append( template.format(**locals()) )
    # ret.append( 'return lt' )
    # make_indent( ret, 2 )
    # return '\n'.join( ret )
    return "    return ''"

  #=========================================================================
  # Helper functions
  #=========================================================================

  def _verilator_name( s, name ):
    # TODO: PyMTL translation should generate dollar-sign-free Verilog source
    # code. Verify that this replacement rule here is not necessary.
    return name.replace('__', '___05F').replace('$', '__024')

  def _pymtl_name_mangle( s, name ):
    return name.replace('.', '_DOT_').replace('[', '_LB_').replace(']', '_RB_')

  def _get_direction( s, port ):
    if isinstance( port, rt.Port ):
      d = port.get_direction()
    elif isinstance( port, rt.Array ):
      d = port.get_sub_type().get_direction()
    else:
      assert False, f"{port} is not a port or array of ports!"
    if d == 'input':
      return 'InPort'
    elif d == 'output':
      return 'OutPort'
    else:
      assert False, f"unrecognized direction {d}!"

  def _get_c_n_dim( s, port ):
    if isinstance( port, rt.Array ):
      return port.get_dim_sizes()
    else:
      return []

  def _get_c_dim( s, port ):
    return "".join( f"[{i}]" for i in s._get_c_n_dim(port) )

  def _get_c_nbits( s, port ):
    if isinstance( port, rt.Array ):
      dtype = port.get_sub_type().get_dtype()
    else:
      dtype = port.get_dtype()
    return dtype.get_length()

  def _gen_ref_write( s, lhs, rhs, nbits, equal='=' ):
    if nbits <= 64:
      return [ '', f"{lhs}[0] {equal} int({rhs})" ]
    else:
      ret = [ '', f'x = {lhs}' ]
      ITEM_BITWIDTH = 32
      num_assigns = (nbits-1)//ITEM_BITWIDTH+1
      for idx in range(num_assigns):
        l = ITEM_BITWIDTH*idx
        r = l+ITEM_BITWIDTH if l+ITEM_BITWIDTH <= nbits else nbits
        ret.append( f"x[{idx}] {equal} int({rhs}[{l}:{r}])" )
      return ret

  def _gen_ref_read( s, lhs, rhs, nbits, equal='=' ):
    if nbits <= 64:
      return [ '', f"{lhs} {equal} {rhs}[0]" ]
    else:
      ret = [ '', f'x = {rhs}' ]
      ITEM_BITWIDTH = 32
      num_assigns = (nbits-1)//ITEM_BITWIDTH+1
      for idx in range(num_assigns):
        l = ITEM_BITWIDTH*idx
        r = l+ITEM_BITWIDTH if l+ITEM_BITWIDTH <= nbits else nbits
        _nbits = r - l
        ret.append( f"{lhs}[{l}:{r}] @= x[{idx}]" )
      return ret

  def _gen_bits_decl( s, nbits ):
    if nbits < 256:
      return f'Bits{nbits}'
    else:
      return f'mk_bits({nbits})'



#=========================================================================
# RTLIRPass.py
#=========================================================================
# Provides the rtlir_getter metadata key.
#
# Author : Peitian Pan
# Date   : May 17, 2020

from pymtl3 import MetadataKey
from pymtl3.passes.BasePass import BasePass


class RTLIRPass( BasePass ):

  #: An RTLIR getter
  #:
  #: Type: ``RTLIRGetter``; output
  rtlir_getter = MetadataKey()



#=========================================================================
# Mamba2020Pass.py
#=========================================================================
# This pass implements trace breaking techniques and supports non-DAG, so
# we don't need the old TraceBreaking pass which only supports DAG
# anymore.
#
# Author : Shunning Jiang
# Date   : Feb 14, 2020

import os
from collections import defaultdict, deque

import py

from pymtl3.datatypes import Bits, is_bitstruct_class
from pymtl3.dsl import MethodPort
from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.extra.pypy import custom_exec
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError

from ..sim.DynamicSchedulePass import kosaraju_scc
from ..sim.SimpleSchedulePass import SimpleSchedulePass, dump_dag
from .HeuristicTopoPass import CountBranchesLoops
from .UnrollSimPass import UnrollSimPass

# _DEBUG = True
_DEBUG = False

class Mamba2020Pass( UnrollSimPass ):

  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    if hasattr( top, "_sched" ):
      raise Exception("Some schedule pass has already been applied!")

    top._sched = PassMetadata()

    # Extract branchiness first
    # Initialize all generated net block to 0 branchiness

    self.meta_block_id = 0
    self.branchiness = { x: 0 for x in top._dag.genblks }
    self.only_loop_at_top = { x: False for x in top._dag.genblks }
    v = CountBranchesLoops()

    # Shunning: since each loop turns into call_assembler_r, a pure-loop
    # update block is basically 0 branchiness and can be inserted anywhere.
    # At the beginning I tried not to put those blocks into any metablock
    # to avoid double call_assembler_r but it just turned out that there
    # is no difference of where you put the call_assembler_r.. Plus,
    # treating loop blocks as normal update block can activate subsequent
    # schedulable 0-branchiness block.

    for blk in top.get_all_update_blocks():
      hostobj = top.get_update_block_host_component( blk )
      if blk in top._dag.blk_greenlet_mapping:
        gblk = top._dag.blk_greenlet_mapping[ blk ]
        self.branchiness[ gblk ], self.only_loop_at_top[ gblk ] = 0, 0
      else:
        self.branchiness[ blk ], self.only_loop_at_top[ blk ] = v.enter( hostobj.get_update_block_info( blk )[-1] )

    self.schedule_ff( top )

    # Reuse simple's flip schedule
    simple = SimpleSchedulePass()
    simple.schedule_posedge_flip( top )

    self.schedule_intra_cycle( top )

    top._sim = PassMetadata()
    self.create_print_line_trace( top )
    self.create_sim_cycle_count( top )
    self.create_lock_unlock_simulation( top )
    top.lock_in_simulation()

    self.create_sim_eval_comb( top )
    self.create_sim_tick( top )
    self.create_sim_reset( top )

  #-----------------------------------------------------------------------
  # compile_meta_block
  #-----------------------------------------------------------------------
  # Util function to compile and trace-break a list of update blocks

  def compile_meta_block( self, blocks ):

    meta_id = self.meta_block_id
    self.meta_block_id += 1

    # Create custom global dict for all blocks inside the meta block
    _globals = { f"blk{i}": b for i, b in enumerate( blocks ) }

    blk_srcs = []
    for i, b in enumerate(blocks):
      # This is a normal update block
      if b in self.branchiness:
        blk_srcs.append( f"blk{i}() # [br {self.branchiness[b]}, loop {int(self.only_loop_at_top[b])}] {b.__name__}" )
      # This is an SCC block which has zero BR and is a loop
      else:
        blk_srcs.append( f"blk{i}() # {b.__name__}" )

    gen_src = f"def meta_block{meta_id}():\n  "
    gen_src += "\n  ".join( blk_srcs )

    # use custom_exec to compile the meta block
    _locals = {}
    custom_exec( py.code.Source( gen_src ).compile(), _globals, _locals )
    ret = _locals[ f'meta_block{meta_id}' ]
    if _DEBUG: print(gen_src)

    # We will use pypyjit.dont_trace_here to compile standalone traces for
    # each meta block
    try:
      from pypyjit import dont_trace_here
      dont_trace_here( 0, False, ret.__code__ )
    except:
      pass

    return ret

  #-----------------------------------------------------------------------
  # schedule_ff
  #-----------------------------------------------------------------------
  # Schedule all update_ff blocks to meta blocks
  # This one is actually easier because we can order them arbitrarily

  def schedule_ff( self, top ):

    top._sched.schedule_ff = schedule = []
    if not top.get_all_update_ff():
      return

    # tuples in ffs: ( branchiness, blk )

    ffs = []
    for x in top.get_all_update_ff():
      # Here we treat loop-only upblk as 0 branchiness
      ffs.append( (0 if self.only_loop_at_top[x] else self.branchiness[x], x) )
    ffs = sorted( ffs, key=lambda x:x[0] )

    # Divide all blks into meta blocks

    # Branchiness factor is the bound of branchiness in a meta block.
    branchiness_factor = 20

    # Block factor is the bound of the number of branchy blocks in a
    # meta block.
    branchy_block_factor = 6

    cur_meta, cur_br, cur_count = [], 0, 0

    for i, (br, blk) in enumerate( ffs ):
      if br == 0:
        cur_meta.append( blk )

      else: # this means the remaining blocks are all branchy
        cur_meta.append( blk )
        cur_br += br
        cur_count += 1

        if cur_br >= branchiness_factor or cur_count >= branchy_block_factor:
          schedule.append( self.compile_meta_block( cur_meta ) )
          cur_br = cur_count = 0
          cur_meta = []

    if cur_meta:
      schedule.append( self.compile_meta_block( cur_meta ) )

  #-----------------------------------------------------------------------
  # schedule_intra_cycle
  #-----------------------------------------------------------------------
  # This is the real entree

  def schedule_intra_cycle( self, top ):

    # Construct the intra-cycle graph based on normal update blocks

    V   = top._dag.final_upblks - top.get_all_update_ff()

    G   = { v: [] for v in V }
    G_T = { v: [] for v in V } # transpose graph

    E = set()
    for (u, v) in top._dag.all_constraints: # u -> v
      if u in V and v in V:
        G  [u].append( v )
        G_T[v].append( u )
        E.add( (u, v) )

    if 'MAMBA_DAG' in os.environ:
      dump_dag( top, V, E )

    # Compute SCC using Kosaraju's algorithm

    SCCs, G_new = kosaraju_scc( G, G_T )

    onces = top.get_all_update_once()
    # This function compiles a SCC block
    scc_id = 0 # global id across all sccs
    def compile_scc( i ):
      nonlocal scc_id

      scc = SCCs[i]

      if len(scc) == 1:
        return list(scc)[0]

      for x in scc:
        if x in onces:
          raise UpblkCyclicError("update_once blocks are not allowed to appear in a cycle. \n - " + \
                          "\n - ".join( [
                            f"{y.__name__} ({'@update_once' if y in onces else '@update'} " \
                            f"in 'top.{repr(top.get_update_block_host_component(y))[2:]}')"
                            for y in scc] ))

      scc_id += 1
      if _DEBUG: print( f"{'='*100}\n SCC{scc_id}\n{'='*100}" )

      # For each non-trivial SCC, we need to figure out a intra-SCC
      # linear schedule that minimizes the time to re-execute this SCC
      # due to value changes. A bad schedule may inefficiently execute
      # the SCC for many times, each of which changes a few signals.
      # The current algorithm iteratively finds the "entry block" of
      # the SCC and expand its adjancent blocks. The implementation is
      # to first find the actual entry point, and then BFS to expand the
      # footprint until all nodes are visited.

      tmp_schedule = []
      Q = deque()

      if scc_pred[i] is None:
        # We start bfs from the block that has the least number of input
        # edges in the SCC
        InD = { v: 0 for v in scc }
        for (u, v) in E: # u -> v
          if u in scc and v in scc:
            InD[ v ] += 1
        Q.append( max(InD, key=InD.get) )

      else:
        # We start bfs with the blocks that are successors of the
        # predecessor scc in the previous SCC-level topological sort.
        pred = set( SCCs[ scc_pred[i] ] )
        # Sort by names for a fixed outcome
        for x in sorted( scc, key = lambda x: x.__name__ ):
          for v in G_T[x]: # find reversed edges point back to pred SCC
            if v in pred:
              Q.append( x )

      # Perform bfs to find a heuristic schedule
      visited = set(Q)
      while Q:
        u = Q.popleft()
        tmp_schedule.append( u )
        for v in G[u]:
          if v in scc and v not in visited:
            Q.append( v )
            visited.add( v )

      variables = set()
      for (u, v) in E:
        # Collect all variables that triggers other blocks in the SCC
        if u in scc and v in scc:
          variables.update( constraint_objs[ (u, v) ] )

      if len(variables) == 0:
        raise UpblkCyclicError("There is a cyclic dependency without involving variables."
                        "Probably a loop that involves blocks that should be update_once:\n{}"\
                        .format(", ".join( [ x.__name__ for x in scc] )))

      # generate a loop for scc
      # Shunning: we just simply loop over the whole SCC block
      # TODO performance optimizations using Mamba techniques within a SCC block

      template = """
from copy import deepcopy
def wrapped_SCC_{0}():
  N = 0
  while True:
    N += 1
    if N > 100:
      raise UpblkCyclicError("Combinational loop detected at runtime in {{{4}}} after 100 iters!")
    {1}
    {3}
    {2}
    # print( "SCC block{0} is executed", N, "times" )
    break
generated_block = wrapped_SCC_{0}
"""

      # clean up non-top variables if top is there. For slices of Bits
      # we directly use the top level wide Bits since Bits clone is
      # rpython code

      final_variables = set()

      for x in sorted( variables, key=repr ):
        w = x.get_top_level_signal()
        if w is x:
          final_variables.add( x )
          continue

        # w is not x
        if issubclass( w._dsl.Type, Bits ):
          if w not in final_variables:
            final_variables.add( w )
        elif is_bitstruct_class( w._dsl.Type ):
          if w not in final_variables:
            final_variables.add( x )
        else:
          final_variables.add( x )

      # also group them by common ancestor to reduce byte code
      # TODO use longest-common-prefix (LCP) algorithms ...

      final_var_host = defaultdict(list)
      for x in final_variables:
        final_var_host[ x.get_host_component() ].append( x )

      # Then, we generate the Python code that saves variables at the
      # beginning of each SCC iteration and the code that checks if the
      # values of those variables have changed
      copy_srcs  = []
      check_srcs = []

      var_id = 0
      for host, var_list in final_var_host.items():
        hostlen = len(repr(host))

        copy_srcs.append( f"host = {host!r}" )
        check_srcs.append( f"host = {host!r}" )

        sub_check_srcs = []

        for var in var_list:
          var_id += 1
          subname = repr(var)[hostlen+1:]
          if issubclass( var._dsl.Type, Bits ):     copy_srcs.append( f"t{var_id}=host.{subname}.clone()" )
          elif is_bitstruct_class( var._dsl.Type ): copy_srcs.append( f"t{var_id}=host.{subname}.clone()" )
          else:                                     copy_srcs.append( f"t{var_id}=deepcopy(host.{subname})" )

          sub_check_srcs.append( f"host.{subname} != t{var_id}" )

        check_srcs.append( f"if { ' or '.join(sub_check_srcs)}: continue" )

      # Divide all blks into meta blocks
      # Branchiness factor is the bound of branchiness in a meta block.
      branchiness_factor = 20
      branchy_block_factor = 6

      num_blks = 0  # sanity check
      cur_meta, cur_br, cur_count = [], 0, 0
      scc_schedule = []

      _globals = { 's': top, 'UpblkCyclicError': UpblkCyclicError }
      blk_srcs = []

      # If there is only 10 blocks, we directly unroll it
      if len(tmp_schedule) < 10:
        blk_srcs = []
        for i, b in enumerate(tmp_schedule):
          blk_srcs.append( f"blk{i}() # [br {self.branchiness[b]}, loop {int(self.only_loop_at_top[b])}] {b.__name__}" )
          _globals[f"blk{i}"] = b # put it into the block's closure

      else:
        for i, blk in enumerate( tmp_schedule ):
          # Same here. If an update block only has top-level loop, br = 0
          br = 0 if self.only_loop_at_top[blk] else self.branchiness[blk]
          if cur_br == 0:
            cur_meta.append( blk )
            cur_br += br
            cur_count += (br > 0)
            if cur_br >= branchiness_factor or cur_count >= branchy_block_factor:
              num_blks += len(cur_meta)
              scc_schedule.append( cur_meta )
              cur_meta, cur_br, cur_count = [], 0, 0 # clear
          else:
            if br == 0:
              # If no branchy block available, directly start a new metablock
              num_blks += len(cur_meta)
              scc_schedule.append( cur_meta )
              cur_meta, cur_br, cur_count = [ blk ], br, (br > 0)
            else:
              cur_meta.append( blk )
              cur_br += br
              cur_count += (br > 0)

              if cur_br + br >= branchiness_factor or cur_count + 1 >= branchy_block_factor:
                num_blks += len(cur_meta)
                scc_schedule.append( cur_meta )
                cur_meta, cur_br, cur_count = [], 0, 0 # clear

        if cur_meta:
          num_blks += len(cur_meta)
          scc_schedule.append( cur_meta )

        assert num_blks == len(tmp_schedule), f"Some blocks are missing during trace breaking of SCC "\
                                              f"({num_blks} compiled, {len(tmp_schedule)} total)"

        blk_srcs = []

        if len(scc_schedule) == 1:
          for i, b in enumerate( scc_schedule[-1] ):
            blk_srcs.append( f"blk{i}() # [br {self.branchiness[b]}, loop {int(self.only_loop_at_top[b])}] {b.__name__}" )
            _globals[ f"blk{i}" ] = b

        else:
          # TODO we might turn all meta blocks before the last one into meta
          # blocks, and directly fold the last block into the main loop
          # for i, meta in enumerate( scc_schedule[:-1] ):
            # b = self.compile_meta_block( meta )
            # blk_srcs.append( f"{b.__name__}()" )
            # _globals[ b.__name__ ] = b

          # for i, b in enumerate( scc_schedule[-1] ):
            # blk_srcs.append( f"blk_of_last_meta{i}() # [br {self.branchiness[b]}, loop {int(self.only_loop_at_top[b])}] {b.__name__}" )
            # _globals[ f"blk_of_last_meta{i}" ] = b

          for i, meta in enumerate( scc_schedule ):
            b = self.compile_meta_block( meta )
            blk_srcs.append( f"{b.__name__}()" )
            _globals[ b.__name__ ] = b


      scc_block_src = template.format( scc_id, "; ".join( copy_srcs ), "\n    ".join( check_srcs ),
                                       '\n    '.join(blk_srcs),
                                       ", ".join( [ x.__name__ for x in scc] ) )

      if _DEBUG: print(scc_block_src, "\n", "="*100 )

      _locals  = {}
      custom_exec(py.code.Source( scc_block_src ).compile(), _globals, _locals)
      return _locals[ 'generated_block' ]

    # Now we generate meta blocks for each SCC and produce final schedule

    constraint_objs = top._dag.constraint_objs

    # Perform topological sort on SCCs

    InD = { i: 0 for i in range(len(SCCs)) }
    nontrivial_sccs   = set()
    trivial_loop_sccs = set()

    for u, vs in G_new.items():
      # Preprocess some sets to mark non-trivial sccs and loop-only sccs
      # for later lookup
      if len(SCCs[u]) > 1:
        nontrivial_sccs.add( u )
      elif self.only_loop_at_top[ list(SCCs[u])[0] ]:
        trivial_loop_sccs.add( u )

      for v in vs:
        InD[ v ] += 1

    # Shunning: reuse this binary search from TraceBreakingPass. Not the
    # most efficient one. Ideally we want to use two heaps or a balanced
    # binary search tree ... TODO

    def insert_sortedlist( arr, key, item ):
      left, right = -1, len(arr)
      while left + 1 < right:
        mid = (left + right) >> 1
        if arr[mid][0] <= key:
          left = mid
        else:
          right = mid
      arr.insert( right, ( key, item ) )

    # scc_pred is for heuristic hamiltonian path ... It records for each
    # scc, in the schedule who is the predecessor that reduce its input
    # degree to zero.
    scc_pred = {}

    Q = []
    cnt = 0

    # Put the graph input nodes into the queue
    for v in range(len(SCCs)):
      if not InD[v]:
        cnt += 1
        scc_pred[v] = None
        if v in nontrivial_sccs or v in trivial_loop_sccs:
          insert_sortedlist( Q, (0, -cnt), v )
        else:
          insert_sortedlist( Q, (self.branchiness[list(SCCs[v])[0]], -cnt), v )

    schedule = []

    # Branchiness factor is the bound of branchiness in a meta block.
    branchiness_factor = 20

    # Block factor is the bound of the number of branchy blocks in a
    # meta block.
    branchy_block_factor = 6

    # refactored code ...
    def expand_node( u ):
      nonlocal cnt
      for v in G_new[u]:
        InD[v] -= 1
        if not InD[v]:
          cnt += 1
          scc_pred[ v ] = u
          # Now we use (br, timestamp) as the key because we want to kind
          # of preserve DFS behavior on top of the branch priority
          # Basically we want to pop in a DFS order such that the variable
          # most recently written can directly feed into the next block
          if v in nontrivial_sccs or v in trivial_loop_sccs:
            insert_sortedlist( Q, (0, -cnt), v )
          else:
            insert_sortedlist( Q, (self.branchiness[list(SCCs[v])[0]], -cnt), v )

    # Run topological sort

    cur_meta = []
    cur_br = cur_count = 0

    while Q:
      if cur_br == 0:
        (br, _), u = Q.pop(0)

        cur_meta.append( compile_scc(u) )
        cur_br += br
        cur_count += (br > 0)

        if cur_br >= branchiness_factor:
          schedule.append( cur_meta )
          cur_meta, cur_br, cur_count = [], 0, 0

      else:
        (br, _), u = Q.pop()

        # If no branchy block available, directly start a new metablock
        if br == 0:
          schedule.append( cur_meta )
          cur_meta, cur_br, cur_count = [], 0, 0

          cur_meta.append( compile_scc(u) )

        # Limit the number of branchiness and number of branchy blocks
        else:
          cur_meta.append( compile_scc(u) )
          cur_br += br
          cur_count += (br > 0)

          if cur_br + br >= branchiness_factor or cur_count + 1 >= branchy_block_factor:
            schedule.append( cur_meta )
            cur_meta, cur_br, cur_count = [], 0, 0

      expand_node( u )

    if cur_meta:
      schedule.append( cur_meta )

    # Put the graph schedule to _sched
    top._sched.update_schedule = []

    # TODO Same as what we want to do for the last block in SCC, we might
    # be able to remove the overhead of the last block's call_assembler_r
    # since the tracing will end anyway. We compile all meta block except
    # for the last one, and directly fold the last meta block into the
    # main loop.
    # We might want to add some heuristics to switch between two modes
    # for i, meta in enumerate( schedule[:-1] ):
      # top._sched.update_schedule.append( self.compile_meta_block( meta ) )
    # for i, b in enumerate( schedule[-1] ):
      # top._sched.update_schedule.append( b )
      # if _DEBUG: print( f"blk_of_last_meta{i}() # [br {self.branchiness[b]}, loop {int(self.only_loop_at_top[b])}] {b.__name__}" )
    if len(schedule) == 1:
      for i, b in enumerate( schedule[0] ):
        top._sched.update_schedule.append( b )
        if _DEBUG: print( f"blk{i}() # [br {self.branchiness[b]}, loop {int(self.only_loop_at_top[b])}] {b.__name__}" )
    else:
      for i, meta in enumerate( schedule ):
        top._sched.update_schedule.append( self.compile_meta_block( meta ) )



#=========================================================================
# elf_test.py
#=========================================================================

import random
import struct

from .. import elf
from ..SparseMemoryImage import SparseMemoryImage

#-------------------------------------------------------------------------
# test_basic
#-------------------------------------------------------------------------

def test_basic( tmpdir ):

  # Create a sparse memory image

  mem_image = SparseMemoryImage()

  section_names = [ ".text", ".data" ]

  for i in range(4):

    section = SparseMemoryImage.Section()
    section.name = section_names[ random.randint(0,1) ]
    section.addr = i * 0x00000200

    data_ints  = [ random.randint(0,1000) for r in range(10) ]
    data_bytes = bytearray()
    for data_int in data_ints:
      data_bytes.extend(struct.pack("<I",data_int))

    section.data = data_bytes

    mem_image.add_section( section )

  # Write the sparse memory image to an ELF file

  with tmpdir.join("elf-test").open('wb') as file_obj:
    elf.elf_writer( mem_image, file_obj )

  # Read the ELF file back into a new sparse memory image

  mem_image_test = None
  with tmpdir.join("elf-test").open('rb') as file_obj:
    mem_image_test = elf.elf_reader( file_obj )

  # Check that the original and new sparse memory images are equal

  assert mem_image == mem_image_test



"""
========================================================================
MemMsg.py
========================================================================
New memory message type implementation.

Author : Shunning Jiang, Yanghui Ou
Date   : Mar 12, 2018
"""
from pymtl3 import *


class MemMsgType:
  READ       = 0
  WRITE      = 1
  # write no-refill
  WRITE_INIT = 2
  AMO_ADD    = 3
  AMO_AND    = 4
  AMO_OR     = 5
  AMO_SWAP   = 6
  AMO_MIN    = 7
  AMO_MINU   = 8
  AMO_MAX    = 9
  AMO_MAXU   = 10
  AMO_XOR    = 11
  LR         = 12 # Load-Reserved
  SC         = 13 # Store-Conditional
  INV        = 14 # Cache invalidation
  FLUSH      = 15 # Cache flush

  str = {
    READ       : "rd",
    WRITE      : "wr",
    WRITE_INIT : "in",
    AMO_ADD    : "ad",
    AMO_AND    : "an",
    AMO_OR     : "or",
    AMO_SWAP   : "sw",
    AMO_MIN    : "mi",
    AMO_MINU   : "mu",
    AMO_MAX    : "mx",
    AMO_MAXU   : "xu",
    AMO_XOR    : "xo",
    LR         : "lr",
    SC         : "sc",
    INV        : "iv",
    FLUSH      : "fl"
  }

def mk_mem_msg( opq, addr, data ):
  return mk_mem_req_msg( opq, addr, data ), mk_mem_resp_msg( opq, data )

def mk_mem_req_msg( o, a, d ):

  @bitstruct
  class MemReqMsg:
    type_  : Bits4
    opaque : mk_bits( o           )
    addr   : mk_bits( a           )
    len    : mk_bits( clog2(d>>3) )
    data   : mk_bits( d           )

    data_nbits = d

    def __str__( self ):
      return "{}:{}:{}:{}:{}".format(
        MemMsgType.str[ int( self.type_ ) ],
        self.opaque,
        self.addr,
        self.len,
        self.data if self.type_ != MemMsgType.READ else " " * ( d//4 ),
      )

  return MemReqMsg

def mk_mem_resp_msg( o, d ):

  @bitstruct
  class MemRespMsg:
    type_  : Bits4
    opaque : mk_bits( o           )
    test   : Bits2
    len    : mk_bits( clog2(d>>3) )
    data   : mk_bits( d           )

    data_nbits = d

    def __str__( self ):
      return "{}:{}:{}:{}:{}".format(
        MemMsgType.str[ int( self.type_ ) ],
        self.opaque,
        self.test,
        self.len,
        self.data if self.type_ != MemMsgType.WRITE else " " * ( d//4 ),
      )
  return MemRespMsg



"""
=========================================================================
 StallCL.py
=========================================================================
 Models random stall

 Author : Shunning Jiang
 Date   : Feb 6, 2020
"""

from random import Random

from pymtl3 import *

# This stall is for testing purpose
# Recv side has a random stall

class StallCL( Component ):

  # ready <==> stall_rand > stall_prob
  @non_blocking( lambda s: s.stall_rgen.random() > s.stall_prob and s.send.rdy() )
  def recv( s, msg ):
    s.send( msg )

  def construct( s, stall_prob=0.5, stall_seed=0x1 ):

    s.send = CallerIfcCL()

    s.stall_prob = stall_prob
    s.stall_rgen = Random( stall_seed ) # Separate randgen for each injector

    s.add_constraints(
      M(s.recv)     == M(s.send),  # pass_through
      M(s.recv.rdy) == M(s.send.rdy),  # pass_through
    )


  def line_trace( s ):
    return f"{s.recv}"



"""
========================================================================
AddDebugSignalPass.py
========================================================================
TODO currently doesn't support array of components because the uniquification
of added signal names of an array of components is very convoluted.
The verilog translation support is limited for two components of the
same class that have different names of debug signals.

Author : Shunning Jiang
Date   : Jul 5, 2020
"""

from pymtl3.dsl import Component, InPort, MetadataKey, OutPort
from pymtl3.passes.BasePass import BasePass


class AddDebugSignalPass( BasePass ):

  debug_pins = MetadataKey(set)

  def __call__( self, top, signal_names ):

    s_signal_names = []
    for name in signal_names:
      assert name.startswith("top.")
      assert '[' not in name, "Currently don't support any array of components"
      s_signal_names.append( f"s{name[3:]}")

    signals = sorted( top.get_all_object_filter( lambda x: repr(x) in s_signal_names ),
                      key=repr )

    for i, signal in enumerate( signals ):
      debug_pin_name = f"debug_{i}"
      last_port = signal
      host = signal.get_host_component()
      while host is not None:
        this_port = OutPort( last_port.get_type() )
        top.add_value_port( host, debug_pin_name, this_port )
        top.add_connection( this_port, last_port )
        last_port = this_port
        host = host.get_parent_object()



'''
==========================================================================
connect_bits2bitstruct.py
==========================================================================
A connect function that connects a bits signal and a bitsrtuct signal that
has the same width.

Author : Yanghui Ou
  Date : Feb 24, 2020
'''
from pymtl3 import Bits, connect
from pymtl3.datatypes.bitstructs import _FIELDS, is_bitstruct_class

#-------------------------------------------------------------------------
# _connect_bits2bitstruct_h
#-------------------------------------------------------------------------
# Helper function for connect_bits2bitstruct.

def _connect_bits2bitstruct_h( field, bits_signal, start ):

  # NOTE: We do not support list of fields at the moment because: 1) there
  # are very few use cases; 2) the pack order of a list of fields is
  # somewhat ambiguous.
  if isinstance( field, list ):
    assert False, 'A list of fields is not supported at the moment.'

  field_type = field.get_type()

  if issubclass( field_type, Bits ):
    stop      = start + field_type.nbits
    cur_slice = slice( start, stop )
    connect( field, bits_signal[ cur_slice ] )
    return stop

  else:
    fields_dict = getattr( field_type, _FIELDS )
    cur_stop    = start
    for fname, _ in reversed( list( fields_dict.items() ) ):
      subfield = getattr( field, fname )
      cur_stop = _connect_bits2bitstruct_h( subfield, bits_signal, cur_stop )
    return cur_stop

#-------------------------------------------------------------------------
# connect_bits2bitstruct
#-------------------------------------------------------------------------

def connect_bits2bitstruct( signal1, signal2 ):
  type1 = signal1.get_type()
  type2 = signal2.get_type()

  if is_bitstruct_class( type1 ):
    assert issubclass( type2, Bits ), \
      f'Connecting {type1.__qualname__} to {type2.__qualname__}: ' \
      f'{type1.__qualname__} is a bitstruct ' \
      f'but {type2.__qualname__} is not a Bits type!'
    bits_signal, bitstruct_signal = signal2, signal1

  elif is_bitstruct_class( type2 ):
    assert issubclass( type1, Bits ), \
      f'Connecting {type1.__qualname__} to {type2.__qualname__}: ' \
      f'{type2.__qualname__} is a bitstruct ' \
      f'but {type1.__qualname__} is not a Bits type!'
    bits_signal, bitstruct_signal = signal1, signal2

  else:
    assert False, 'Can only connect a bitstruct wire to bits wire'

  bw1 = type1.nbits
  bw2 = type2.nbits
  assert bw1 == bw2, 'Bitwidth mismatch! ' \
    f'{type1.__qualname__} is {bw1}-bit ' \
    f'but {type2.__qualname__} is {bw2}-bit.'

  # Connect field to corresponding slice
  stop = _connect_bits2bitstruct_h( bitstruct_signal, bits_signal, 0 )
  assert stop == bw1, 'Sanity check failed! Might be a bug. ' \
    'Please create a Github issue.'



"""
-------------------------------------------------------------------------
Library of RTL queues
-------------------------------------------------------------------------

Author : Yanghui Ou
  Date : Feb 22, 2021
"""


from pymtl3 import *
from pymtl3.stdlib.basic_rtl import Mux, RegisterFile

from .ifcs import RecvIfcRTL, SendIfcRTL

#-------------------------------------------------------------------------
# NormalQueue1EntryRTL
#-------------------------------------------------------------------------

class NormalQueue1EntryRTL( Component ):

  def construct( s, EntryType ):

    # Interface

    s.recv  = RecvIfcRTL( EntryType )
    s.send  = SendIfcRTL( EntryType )
    s.count = OutPort()

    # Components

    s.full  = Wire()
    s.entry = Wire( EntryType )

    # Logic

    s.count //= s.full

    s.send.msg //= s.entry
    s.send.val //= s.full
    s.recv.rdy //= lambda: ~s.full

    @update_ff
    def ff_normal1():
      if s.reset:
        s.full <<= 0
      else:
        s.full <<= (s.recv.val & ~s.full) | (s.full & ~s.send.rdy)

      if s.recv.val & ~s.full:
        s.entry <<= s.recv.msg

  def line_trace( s ):
    return f"{s.recv}({s.full}){s.send}"

#-------------------------------------------------------------------------
# Dpath and Ctrl for NormalQueueRTL
#-------------------------------------------------------------------------

class NormalQueueDpathRTL( Component ):

  def construct( s, EntryType, num_entries=2 ):
    assert num_entries >= 2

    # Interface

    s.recv_msg = InPort( EntryType )
    s.send_msg = OutPort( EntryType )

    s.wen   = InPort()
    s.waddr = InPort( clog2( num_entries ) )
    s.raddr = InPort( clog2( num_entries ) )

    # Component

    s.rf = m = RegisterFile( EntryType, num_entries )
    m.raddr[0] //= s.raddr
    m.rdata[0] //= s.send_msg
    m.wen[0]   //= s.wen
    m.waddr[0] //= s.waddr
    m.wdata[0] //= s.recv_msg

class NormalQueueCtrlRTL( Component ):

  def construct( s, num_entries=2 ):
    assert num_entries >= 2

    # Constants

    addr_nbits  = clog2( num_entries   )
    count_nbits = clog2( num_entries+1 )

    # Interface

    s.recv_val = InPort()
    s.recv_rdy = OutPort()
    s.send_val = OutPort()
    s.send_rdy = InPort()

    s.count = OutPort( count_nbits )
    s.wen   = OutPort()
    s.waddr = OutPort( addr_nbits )
    s.raddr = OutPort( addr_nbits )

    # Registers

    s.head = Wire( addr_nbits )
    s.tail = Wire( addr_nbits )

    # Wires

    s.recv_xfer  = Wire()
    s.send_xfer  = Wire()

    # Connections

    s.wen   //= s.recv_xfer
    s.waddr //= s.tail
    s.raddr //= s.head

    s.recv_rdy  //= lambda: s.count < num_entries
    s.send_val  //= lambda: s.count > 0

    s.recv_xfer //= lambda: s.recv_val & s.recv_rdy
    s.send_xfer //= lambda: s.send_val & s.send_rdy

    @update_ff
    def up_reg():

      if s.reset:
        s.head  <<= 0
        s.tail  <<= 0
        s.count <<= 0

      else:
        if s.recv_xfer:
          s.tail <<= s.tail + 1 if ( s.tail < num_entries - 1 ) else 0

        if s.send_xfer:
          s.head <<= s.head + 1 if ( s.head < num_entries -1 ) else 0

        if s.recv_xfer & ~s.send_xfer:
          s.count <<= s.count + 1
        elif ~s.recv_xfer & s.send_xfer:
          s.count <<= s.count - 1

#-------------------------------------------------------------------------
# NormalQueueRTL
#-------------------------------------------------------------------------

class NormalQueueRTL( Component ):

  def construct( s, EntryType, num_entries=2 ):

    # Interface

    s.recv  = RecvIfcRTL( EntryType )
    s.send  = SendIfcRTL( EntryType )
    s.count = OutPort( clog2( num_entries+1 ) )

    # Components

    assert num_entries > 0

    if num_entries == 1:
      s.q = NormalQueue1EntryRTL( EntryType )
      s.recv  //= s.q.recv
      s.send  //= s.q.send
      s.count //= s.q.count

    else:
      s.ctrl  = NormalQueueCtrlRTL ( num_entries )
      s.dpath = NormalQueueDpathRTL( EntryType, num_entries )

      # Connect ctrl to data path

      s.ctrl.wen   //= s.dpath.wen
      s.ctrl.waddr //= s.dpath.waddr
      s.ctrl.raddr //= s.dpath.raddr

      # Connect to interface

      s.recv.val //= s.ctrl.recv_val
      s.recv.rdy //= s.ctrl.recv_rdy
      s.recv.msg //= s.dpath.recv_msg

      s.send.val //= s.ctrl.send_val
      s.send.rdy //= s.ctrl.send_rdy
      s.send.msg //= s.dpath.send_msg
      s.count   //= s.ctrl.count

  # Line trace

  def line_trace( s ):
    return f"{s.recv}({s.count}){s.send}"

#-------------------------------------------------------------------------
# PipeQueue1EntryRTL
#-------------------------------------------------------------------------

class PipeQueue1EntryRTL( Component ):

  def construct( s, EntryType ):

    # Interface

    s.recv  = RecvIfcRTL( EntryType )
    s.send  = SendIfcRTL( EntryType )
    s.count = OutPort()

    # Components

    s.full  = Wire()
    s.entry = Wire( EntryType )

    # Logic

    s.count //= s.full

    s.send.msg //= s.entry
    s.send.val //= s.full

    # If send is rdy, either the entry will be sent out to free up space
    # for recv, or entry is already available for send. Then if not full
    # entry can always buffer up a message. rdy path is elongated
    s.recv.rdy //= lambda: s.send.rdy | ~s.full

    @update_ff
    def ff_pipe1():
      if s.reset:
        s.full <<= 0
      else:
        # The pipe queue is full if in this cycle it cannot receive a
        # message due to back pressure and the entry is already full.
        # Otherwise it is not full and it becomes full only if there is
        # a valid incoming message.
        s.full <<= ~s.recv.rdy | s.recv.val

      # AND rdy and val to buffer the incoming message
      if s.recv.rdy & s.recv.val:
        s.entry <<= s.recv.msg

  def line_trace( s ):
    return f"{s.recv}({s.full}){s.send}"

#-------------------------------------------------------------------------
# Ctrl for PipeQueue
#-------------------------------------------------------------------------

class PipeQueueCtrlRTL( Component ):

  def construct( s, num_entries=2 ):
    assert num_entries >= 2

    # Constants

    addr_nbits  = clog2( num_entries   )
    count_nbits = clog2( num_entries+1 )

    # Interface

    s.recv_val = InPort ()
    s.recv_rdy = OutPort()
    s.send_val = OutPort()
    s.send_rdy = InPort ()
    s.count    = OutPort( count_nbits )

    s.wen      = OutPort()
    s.waddr    = OutPort( addr_nbits )
    s.raddr    = OutPort( addr_nbits )

    # Registers

    s.head = Wire( addr_nbits )
    s.tail = Wire( addr_nbits )

    # Wires

    s.recv_xfer  = Wire()
    s.send_xfer  = Wire()

    # Connections

    s.wen   //= s.recv_xfer
    s.waddr //= s.tail
    s.raddr //= s.head

    s.send_val //= lambda: s.count > 0
    s.recv_rdy //= lambda: ( s.count < num_entries ) | s.send_rdy

    s.recv_xfer //= lambda: s.recv_val & s.recv_rdy
    s.send_xfer //= lambda: s.send_val & s.send_rdy

    @update_ff
    def up_reg():

      if s.reset:
        s.head  <<= 0
        s.tail  <<= 0
        s.count <<= 0

      else:
        if s.recv_xfer:
          s.tail <<= s.tail + 1 if ( s.tail < num_entries - 1 ) else 0

        if s.send_xfer:
          s.head <<= s.head + 1 if ( s.head < num_entries -1 ) else 0

        if s.recv_xfer & ~s.send_xfer:
          s.count <<= s.count + 1
        elif ~s.recv_xfer & s.send_xfer:
          s.count <<= s.count - 1

#-------------------------------------------------------------------------
# PipeQueueRTL
#-------------------------------------------------------------------------

class PipeQueueRTL( Component ):

  def construct( s, EntryType, num_entries=2 ):

    # Interface

    s.recv  = RecvIfcRTL( EntryType )
    s.send  = SendIfcRTL( EntryType )
    s.count = OutPort( clog2( num_entries+1 ) )

    # Components

    assert num_entries > 0
    if num_entries == 1:
      s.q = PipeQueue1EntryRTL( EntryType )
      s.recv  //= s.q.recv
      s.send  //= s.q.send
      s.count //= s.q.count

    else:
      s.ctrl  = PipeQueueCtrlRTL ( num_entries )
      s.dpath = NormalQueueDpathRTL( EntryType, num_entries )

      # Connect ctrl to data path

      s.ctrl.wen   //= s.dpath.wen
      s.ctrl.waddr //= s.dpath.waddr
      s.ctrl.raddr //= s.dpath.raddr

      # Connect to interface

      s.recv.val //= s.ctrl.recv_val
      s.recv.rdy //= s.ctrl.recv_rdy
      s.send.val //= s.ctrl.send_val
      s.send.rdy //= s.ctrl.send_rdy
      s.count    //= s.ctrl.count
      s.recv.msg //= s.dpath.recv_msg
      s.send.msg //= s.dpath.send_msg

  # Line trace

  def line_trace( s ):
    return f"{s.recv}({s.count}){s.send}"

#-------------------------------------------------------------------------
# BypassQueue1EntryRTL
#-------------------------------------------------------------------------

class BypassQueue1EntryRTL( Component ):

  def construct( s, EntryType ):

    # Interface

    s.recv  = RecvIfcRTL( EntryType )
    s.send  = SendIfcRTL( EntryType )
    s.count = OutPort()

    # Components

    s.full  = Wire()
    s.entry = Wire( EntryType )

    s.bypass_mux = m = Mux( EntryType, 2 )
    m.in_[0] //= s.recv.msg
    m.in_[1] //= s.entry
    m.out    //= s.send.msg
    m.sel    //= s.full

    # Logic

    s.count //= s.full

    s.send.val //= lambda: s.full | s.recv.val
    s.recv.rdy //= lambda: ~s.full

    @update_ff
    def ff_bypass1():
      if s.reset:
        s.full <<= 0
      else:
        s.full <<= ~s.send.rdy & (s.full | s.recv.val)

      # buffer the incoming message if we cannot directly send it out
      if ~s.send.rdy & ~s.full & s.recv.val:
        s.entry <<= s.recv.msg

  def line_trace( s ):
    return f"{s.recv}({s.full}){s.send}"
#-------------------------------------------------------------------------
# Ctrl and Dpath for BypassQueue
#-------------------------------------------------------------------------

class BypassQueueDpathRTL( Component ):

  def construct( s, EntryType, num_entries=2 ):
    assert num_entries >= 2

    # Interface

    s.recv_msg = InPort( EntryType )
    s.send_msg = OutPort( EntryType )

    s.wen     = InPort()
    s.waddr   = InPort( clog2( num_entries ) )
    s.raddr   = InPort( clog2( num_entries ) )
    s.mux_sel = InPort()

    # Component

    s.rf = m = RegisterFile( EntryType, num_entries )
    m.raddr[0] //= s.raddr
    m.wen[0]   //= s.wen
    m.waddr[0] //= s.waddr
    m.wdata[0] //= s.recv_msg

    s.mux = m = Mux( EntryType, 2 )
    m.sel    //= s.mux_sel
    m.in_[0] //= s.rf.rdata[0]
    m.in_[1] //= s.recv_msg
    m.out    //= s.send_msg

class BypassQueueCtrlRTL( Component ):

  def construct( s, num_entries=2 ):
    assert num_entries >= 2

    # Constants

    addr_nbits  = clog2( num_entries   )
    count_nbits = clog2( num_entries+1 )

    last_idx = num_entries-1

    # Interface

    s.recv_val = InPort ()
    s.recv_rdy = OutPort()
    s.send_val = OutPort()
    s.send_rdy = InPort()
    s.count    = OutPort( count_nbits )

    s.wen     = OutPort()
    s.waddr   = OutPort( addr_nbits )
    s.raddr   = OutPort( addr_nbits )
    s.mux_sel = OutPort()

    # Registers

    s.head = Wire( addr_nbits )
    s.tail = Wire( addr_nbits )

    # Wires

    s.recv_xfer  = Wire()
    s.send_xfer  = Wire()

    # Connections

    s.wen   //= s.recv_xfer
    s.waddr //= s.tail
    s.raddr //= s.head

    s.recv_rdy //= lambda: s.count < num_entries
    s.send_val //= lambda: (s.count > 0) | s.recv_val

    s.mux_sel //= lambda: s.count == 0

    s.recv_xfer //= lambda: s.recv_val & s.recv_rdy
    s.send_xfer //= lambda: s.send_val & s.send_rdy

    @update_ff
    def up_reg():

      if s.reset:
        s.head  <<= 0
        s.tail  <<= 0
        s.count <<= 0

      else:
        if s.recv_xfer:
          s.tail <<= s.tail + 1 if ( s.tail < num_entries - 1 ) else 0

        if s.send_xfer:
          s.head <<= s.head + 1 if ( s.head < num_entries -1 ) else 0

        if s.recv_xfer & ~s.send_xfer:
          s.count <<= s.count + 1
        if ~s.recv_xfer & s.send_xfer:
          s.count <<= s.count - 1

#-------------------------------------------------------------------------
# BypassQueueRTL
#-------------------------------------------------------------------------

class BypassQueueRTL( Component ):

  def construct( s, EntryType, num_entries=2 ):

    # Interface

    s.recv  = RecvIfcRTL( EntryType )
    s.send  = SendIfcRTL( EntryType )
    s.count = OutPort( clog2( num_entries+1 ) )

    # Components

    assert num_entries > 0
    if num_entries == 1:
      s.q = BypassQueue1EntryRTL( EntryType )
      s.recv  //= s.q.recv
      s.send  //= s.q.send
      s.count //= s.q.count

    else:
      s.ctrl  = BypassQueueCtrlRTL ( num_entries )
      s.dpath = BypassQueueDpathRTL( EntryType, num_entries )

      # Connect ctrl to data path

      s.ctrl.wen     //= s.dpath.wen
      s.ctrl.waddr   //= s.dpath.waddr
      s.ctrl.raddr   //= s.dpath.raddr
      s.ctrl.mux_sel //= s.dpath.mux_sel

      # Connect to interface

      s.recv.val //= s.ctrl.recv_val
      s.recv.rdy //= s.ctrl.recv_rdy
      s.send.val //= s.ctrl.send_val
      s.send.rdy //= s.ctrl.send_rdy
      s.count    //= s.ctrl.count
      s.recv.msg //= s.dpath.recv_msg
      s.send.msg //= s.dpath.send_msg

  # Line trace

  def line_trace( s ):
    return f"{s.recv}({s.count}){s.send}"




from .test_helpers import (
    RunTestVectorSimError,
    TestVectorSimulator,
    config_model_with_cmdline_opts,
    mk_test_case_table,
    run_sim,
    run_test_vector_sim,
)
from .test_masters import TestMasterCL
from .test_sinks import TestSinkCL
from .test_srcs import TestSrcCL



#=========================================================================
# YosysVerilatorImportPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 14, 2019
"""Provide a pass that imports arbitrary SystemVerilog modules."""
from pymtl3.passes.backends.verilog import VerilogVerilatorImportPass
from pymtl3.passes.backends.yosys.util.utility import gen_mapped_ports


class YosysVerilatorImportPass( VerilogVerilatorImportPass ):

  #-----------------------------------------------------------------------
  # Backend-specific methods
  #-----------------------------------------------------------------------

  @staticmethod
  def get_backend_name():
    return "yosys"

  @staticmethod
  def get_translation_pass():
    from pymtl3.passes.backends.yosys.translation.YosysTranslationPass import (
        YosysTranslationPass,
    )
    return YosysTranslationPass

  @staticmethod
  def get_gen_mapped_port():
    return gen_mapped_ports

  #-----------------------------------------------------------------------
  # Customize assignments in the Python wrapper
  #-----------------------------------------------------------------------

  def gen_port_array_input( s, lhs, rhs, pnames, dtype, index, n_dim, symbols ):
    if not n_dim:
      return s.gen_port_input( lhs, rhs, pnames, dtype, symbols )
    else:
      set_comb, structs = [], []
      for idx in range( n_dim[0] ):
        _lhs = f"{lhs}__{i}"
        if index == 0:
          _rhs = f"{rhs}[{idx}]"
          _index = index-1
        else:
          _rhs = rhs
          _index = index
        _set_comb, _structs = s.gen_port_array_input( _lhs, _rhs, pnames, dtype, _index, n_dim[1:], symbols )
        set_comb += _set_comb
        structs  += _structs
      return set_comb, structs

  def gen_port_array_output( s, lhs, pnames, rhs, dtype, index, n_dim, symbols ):
    if not n_dim:
      return s.gen_port_output( lhs, pnames, rhs, dtype, symbols )
    else:
      set_comb, structs = [], []
      for idx in range( n_dim[0] ):
        if index == 0:
          _lhs = f"{lhs}[{idx}]"
          _index = index-1
        else:
          _lhs = lhs
          _index = index
        _rhs = f"{rhs}__{i}"
        _set_comb, _structs = s.gen_port_array_output( _lhs, pnames, _rhs, dtype, _index, n_dim[1:], symbols )
        set_comb += _set_comb
        structs  += _structs
      return set_comb, structs



"""
==========================================================================
 master_minion_ifc.py
==========================================================================
Master/minion send/recv interface implementations at CL and RTL.
 Author: Shunning Jiang
   Date: Jan 28, 2020
"""
from pymtl3 import *

from .send_recv_ifcs import RecvIfcRTL, SendIfcRTL

#-------------------------------------------------------------------------
# CL interfaces
#-------------------------------------------------------------------------

# There is no custom connect method in CL ifcs.
# For MasterCL-MinionRTL and MasterRTL-MinionCL connections, we just need
# to connect by name and leverage the custom connect method of the nested
# Send/RecvIfc.

class MasterIfcCL( Interface ):
  def construct( s, ReqType, RespType, resp=None, resp_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CallerIfcCL( Type=ReqType )
    s.resp = CalleeIfcCL( Type=RespType, method=resp, rdy=resp_rdy )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

class MinionIfcCL( Interface ):
  def construct( s, ReqType, RespType, req=None, req_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CalleeIfcCL( Type=ReqType, method=req, rdy=req_rdy )
    s.resp = CallerIfcCL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

#-------------------------------------------------------------------------
# RTL interfaces
#-------------------------------------------------------------------------
# There is no custom connect method in CL ifcs.
# For MasterCL-MinionRTL and MasterRTL-MinionCL connections, we just need
# to connect by name and leverage the custom connect method of the nested
# Send/RecvIfc.

class MasterIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = SendIfcRTL( Type=ReqType )
    s.resp = RecvIfcRTL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"

class MinionIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = RecvIfcRTL( Type=ReqType )
    s.resp = SendIfcRTL( Type=RespType )
  def __str__( s ):
    return f"{s.req}|{s.resp}"



#=========================================================================
# SparseMemoryImage_test.py
#=========================================================================

import copy
import random
import struct

from ..SparseMemoryImage import SparseMemoryImage

#-------------------------------------------------------------------------
# test_sections
#-------------------------------------------------------------------------

def test_sections():

  # The sparse memory image we will be testing

  mem_image = SparseMemoryImage()

  # Create first bytearray with some data

  data_ints_a  = [ random.randint(0,1000) for r in range(10) ]
  data_bytes_a = bytearray()
  for data_int in data_ints_a:
    data_bytes_a.extend(struct.pack("<I",data_int))

  # Create a second section section and add it to the sparse memory image

  section_a = SparseMemoryImage.Section( ".text", 0x1000, data_bytes_a )
  mem_image.add_section( section_a )

  # Create a second bytearray with some data

  data_ints_b  = [ random.randint(0,1000) for r in range(10) ]
  data_bytes_b = bytearray()
  for data_ints in data_ints_b:
    data_bytes_b.extend(struct.pack("<I",data_int))

  # Create a second section section and add it to the sparse memory
  # image. Use the alternative syntax for adding a section.

  section_b = SparseMemoryImage.Section( ".data", 0x2000, data_bytes_b )
  mem_image.add_section( ".data", 0x2000, data_bytes_b )

  # Retrieve and check both sections

  section_a_test = mem_image.get_section( ".text" )
  assert section_a_test == section_a

  section_b_test = mem_image.get_section( ".data" )
  assert section_b_test == section_b

  # Retrieve sections as a list

  sections_test = mem_image.get_sections()
  assert sections_test == [ section_a_test, section_b_test ]

#-------------------------------------------------------------------------
# test_equality
#-------------------------------------------------------------------------

def test_equality():

  # The sparse memory image we will be testing

  mem_image = SparseMemoryImage()

  # Create first bytearray with some data

  data_ints_a  = [ random.randint(0,1000) for r in range(10) ]
  data_bytes_a = bytearray()
  for data_int in data_ints_a:
    data_bytes_a.extend(struct.pack("<I",data_int))

  # Create a second section section and add it to the sparse memory image

  section_a = SparseMemoryImage.Section( ".text", 0x1000, data_bytes_a )
  mem_image.add_section( section_a )

  # Create a second bytearray with some data

  data_ints_b  = [ random.randint(0,1000) for r in range(10) ]
  data_bytes_b = bytearray()
  for data_ints in data_ints_b:
    data_bytes_b.extend(struct.pack("<I",data_int))

  # Create a second section section and add it to the sparse memory image

  section_b = SparseMemoryImage.Section( ".data", 0x2000, data_bytes_b )
  mem_image.add_section( section_b )

  # Create a copy

  mem_image_copy = copy.deepcopy( mem_image )

  # Check two images are equal

  assert mem_image == mem_image_copy

  # Add another section to the copy

  section_c = SparseMemoryImage.Section( ".test", 0x3000, bytearray(b"\x01\x02") )
  mem_image_copy.add_section( section_c )

  # Check two images are no longer equal

  assert mem_image != mem_image_copy

#-------------------------------------------------------------------------
# test_symbols
#-------------------------------------------------------------------------

def test_symbols():

  # The sparse memory image we will be testing

  mem_image = SparseMemoryImage()

  # Add some symbols

  mem_image.add_symbol( "func_a", 0x0002000 )
  mem_image.add_symbol( "func_b", 0x0003000 )
  mem_image.add_symbol( "var_a",  0x0011000 )
  mem_image.add_symbol( "var_b",  0x0011000 )

  # Check symbols

  assert mem_image.get_symbol( "func_a" ) == 0x0002000
  assert mem_image.get_symbol( "func_b" ) == 0x0003000
  assert mem_image.get_symbol( "var_a"  ) == 0x0011000
  assert mem_image.get_symbol( "var_b"  ) == 0x0011000



#=========================================================================
# MetadataKey.py
#=========================================================================
# Provides MetadataKey class.
#
# Author : Peitian Pan
# Date   : Mar 16, 2020

class MetadataKey:
  def __init__( self, T=None ):
    assert T is None or isinstance( T, type ), f"The given parameter {T} is not a type."
    self.type = T

  def check_value( self, value ):
    if self.type is not None and not isinstance( value, self.type ):
      raise TypeError( f"Value {value} is of type {type(value)} and cannot be assigned "
                       f"to this MetadataKey that enforces type {self.type}." )



#=========================================================================
# BehavioralRTLIRTypeCheckL1Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 20, 2019
"""Provide L1 behavioral RTLIR type check pass."""

import copy
import math
from collections import deque
from contextlib import contextmanager

from pymtl3 import MetadataKey
from pymtl3.datatypes import Bits32, mk_bits
from pymtl3.passes.rtlir.errors import PyMTLTypeError, RTLIRConversionError
from pymtl3.passes.rtlir.RTLIRPass import RTLIRPass
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from . import BehavioralRTLIR as bir
from .BehavioralRTLIR import BaseBehavioralRTLIR
from .BehavioralRTLIRGenL1Pass import BehavioralRTLIRGenL1Pass


class BehavioralRTLIRTypeCheckL1Pass( RTLIRPass ):

  # Pass metadata

  #: A dictionary that maps free variable names to (object, RTLIRType)
  #:
  #: Type: ``dict``; output
  rtlir_freevars = MetadataKey()

  #: A set of variable names that are accessed in the upblks
  #:
  #: Type: ``set(str)``; output
  rtlir_accessed = MetadataKey()

  def __init__( s, translation_top ):
    c = s.__class__
    s.tr_top = translation_top
    if not translation_top.has_metadata( c.rtlir_getter ):
      translation_top.set_metadata( c.rtlir_getter, rt.RTLIRGetter(cache=True) )

  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    c = s.__class__

    rtlir_freevars = {}
    rtlir_accessed = set()

    m.set_metadata( BehavioralRTLIRTypeCheckL1Pass.rtlir_freevars, rtlir_freevars )
    m.set_metadata( BehavioralRTLIRTypeCheckL1Pass.rtlir_accessed, rtlir_accessed )

    type_checker = s.get_visitor_class()(
      m,
      rtlir_freevars,
      rtlir_accessed,
      s.tr_top.get_metadata( c.rtlir_getter ),
    )

    rtlir_upblks = m.get_metadata( BehavioralRTLIRGenL1Pass.rtlir_upblks )

    for blk in m.get_update_block_order():
      type_checker.enter( blk, rtlir_upblks[ blk ] )

  #-------------------------------------------------------------------------
  # Type checker
  #-------------------------------------------------------------------------

  def get_visitor_class( s ):
    return BehavioralRTLIRTypeCheckVisitorL1

class BehavioralRTLIRTypeCheckVisitorL1( bir.BehavioralRTLIRNodeVisitor ):
  def __init__( s, component, freevars, accessed, rtlir_getter ):
    # Freevars of the same component is probably capture in the same
    # closure/global variable, so we set up RTLIR getter for each apply
    s.rtlir_getter = rtlir_getter
    s.component = component
    s.freevars = freevars
    s.accessed = accessed
    s.type_expect = {}

    # The enforcer is called with a context and an AST node so that it can
    # enforce every implicitly typed node in the subtree to have the type
    # of the given context. Note that after one enforcement, the _is_explicit
    # flag would be set to True.
    try:
      s.enforcer = s.get_enforce_visitor()( component )
    except TypeError:
      pass

    lhs_types = ( rt.Port, rt.Wire, rt.NetWire )
    index_types = ( rt.Port, rt.Wire, rt.Array )
    slice_types = ( rt.Port, rt.Wire )
    s.type_expect[ 'Assign' ] = (
      ( 'targets', lhs_types, 'lhs of assignment must be a signal!' ),
      ( 'value',   rt.Signal, 'rhs of assignment should be signal or const!' ),
    )
    s.type_expect[ 'ZeroExt' ] = (
      ( 'value',   rt.Signal, 'extension only applies to signals!' ),
    )
    s.type_expect[ 'SignExt' ] = (
      ( 'value',   rt.Signal, 'extension only applies to signals!' ),
    )
    s.type_expect[ 'Reduce' ] = (
      ( 'value',   rt.Signal, 'reduce only applies on a signal!' ),
    )
    s.type_expect[ 'SizeCast' ] = (
      ( 'value',   rt.Signal, 'size casting only applies to signals/consts!' ),
    )
    s.type_expect[ 'Attribute' ] = (
      ( 'value',   rt.Component, 'the base of an attribute must be a component!' ),
    )
    s.type_expect[ 'Index' ] = (
      ( 'idx',     rt.Signal, 'index must be a signal or constant expression!' ),
      ( 'value',   index_types, 'the base of an index must be an array or signal!' ),
    )
    s.type_expect[ 'Slice' ] = (
      ( 'value',   slice_types, 'the base of a slice must be a signal!' ),
      ( 'lower',   rt.Signal, 'upper of slice must be a constant expression!' ),
      ( 'upper',   rt.Signal, 'lower of slice must be a constant expression!' ),
    )

  def enter( s, blk, rtlir ):
    """ entry point for RTLIR type checking """
    s.blk     = blk

    # s.globals contains a dict of the global namespace of the module where
    # blk was defined
    s.globals = blk.__globals__

    # s.closure contains the free variables defined in an enclosing scope.
    # Basically this is the model instance s.
    s.closure = {}

    for i, var in enumerate( blk.__code__.co_freevars ):
      try:
        s.closure[ var ] = blk.__closure__[ i ].cell_contents
      except ValueError:
        pass
    s.visit( rtlir )

  def get_enforce_visitor( s ):
    return BehavioralRTLIRTypeEnforcerL1

  # Override the default visit()
  def visit( s, node ):
    # not node._is_explicit: is the parent node allowed to re-interpret this
    # node's bitwidth without truncation?
    node_name = node.__class__.__name__
    func = getattr( s, f'visit_{node_name}', s.generic_visit )

    if node_name == 'For':
      # First visit (type check) all child nodes
      for field, value in vars(node).items():
        # Special case For because we use context depedent types
        # for the loop index
        if field == 'body': continue

        if isinstance( value, BaseBehavioralRTLIR ):
          s.visit( value )
        elif isinstance( value, list ):
          for item in value:
            if isinstance( item, BaseBehavioralRTLIR ):
              s.visit( item )
    else:
      # First visit (type check) all child nodes
      for field, value in vars(node).items():
        if isinstance( value, BaseBehavioralRTLIR ):
          s.visit( value )
        elif isinstance( value, list ):
          for item in value:
            if isinstance( item, BaseBehavioralRTLIR ):
              s.visit( item )

    # Then verify that all child nodes have desired types
    try:
      node_vars = vars(node)
      # Check the expected types of child nodes
      for field, target_type, error_msg in s.type_expect[node_name]:
        value = node_vars[ field ]

        if isinstance( value, list ):
          for v in value:
            if not isinstance( v.Type, target_type ):
              raise PyMTLTypeError( s.blk, node.ast, error_msg )
        else:
          if not isinstance( value.Type, target_type ):
            raise PyMTLTypeError( s.blk, node.ast, error_msg )
    except PyMTLTypeError:
      raise
    except Exception:
      # This node does not require type checking on child nodes
      pass

    # Finally call the type check function
    func( node )

  # Override the default generic_visit()
  def generic_visit( s, node ):
    node.Type = None
    # Is the parent node allowed to re-interpret the type of this node?
    node._is_explicit = True
    # Has this node re-interpreted the type of its child node?
    node._has_reinterpreted = False

  def _visit_Assign_single_target( s, node, target, i ):
    lhs_type = target.Type.get_dtype()
    rhs_type = node.value.Type.get_dtype()

    # At L1 it's always signal assignment
    is_rhs_reinterpretable = not node.value._is_explicit
    if is_rhs_reinterpretable and ((not lhs_type(rhs_type)) or (rhs_type != lhs_type)):
      s.enforcer.enter( s.blk, target.Type, node.value )

    rhs_type = node.value.Type.get_dtype()
    # Weak type checking (agreeable types)
    if not lhs_type( rhs_type ):
      raise PyMTLTypeError( s.blk, node.ast,
        f'Unagreeable types between LHS and RHS (LHS target#{i+1} of {lhs_type} vs {rhs_type})!' )

    # Strong type checking (same type)
    if rhs_type != lhs_type:
      raise PyMTLTypeError( s.blk, node.ast,
        f'LHS and RHS of assignment should have the same type (LHS target#{i+1} of {lhs_type} vs {rhs_type})!' )

  def visit_Assign( s, node ):
    # RHS should have the same type as any of LHS
    for i, target in enumerate( node.targets ):
      s._visit_Assign_single_target( node, target, i )

    node.Type = None
    node._is_explicit = True

  def visit_FreeVar( s, node ):
    try:
      t = s.rtlir_getter.get_rtlir( node.obj )
    except RTLIRConversionError as e:
      raise PyMTLTypeError(s.blk, node.ast,
        f'{node.name} cannot be converted into a valid RTLIR object!' )

    if isinstance( t, rt.Const ) and isinstance( t.get_dtype(), rdt.Vector ):
      node._value = int(node.obj)
    node.Type = t
    node._is_explicit = not isinstance(node.obj, int)

    if node.name not in s.freevars:
      s.freevars[ node.name ] = ( node.obj, t )

  def visit_Base( s, node ):
    # Mark this node as having type rt.Component
    # In L1 the `s` top component is the only possible base
    node.Type = s.rtlir_getter.get_rtlir( node.base )
    node._is_explicit = True
    if not isinstance( node.Type, rt.Component ):
      raise PyMTLTypeError( s.blk, node.ast,
        f'{node} is not a rt.Component!' )

  def visit_Number( s, node ):
    # By default, number literals have the minimal bitwidth that can
    # hold its value without truncation.
    node.Type = s.rtlir_getter.get_rtlir( node.value )
    node._value = int(node.value)
    node._is_explicit = False

  def visit_Concat( s, node ):
    nbits = 0
    for child in node.values:
      if not isinstance(child.Type, rt.Signal):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{child} is not a signal!' )
      nbits += child.Type.get_dtype().get_length()
    node.Type = rt.NetWire( rdt.Vector( nbits ) )
    node._is_explicit = True

  def visit_ZeroExt( s, node ):
    new_nbits = node.nbits
    child_type = node.value.Type
    old_nbits = child_type.get_dtype().get_length()
    if new_nbits < old_nbits:
      raise PyMTLTypeError( s.blk, node.ast,
        f'the target bitwidth {new_nbits} is less than the bitwidth of the operand ({old_nbits})!' )
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( new_nbits )
    node._is_explicit = True

  def visit_SignExt( s, node ):
    new_nbits = node.nbits
    child_type = node.value.Type
    old_nbits = child_type.get_dtype().get_length()
    if new_nbits < old_nbits:
      raise PyMTLTypeError( s.blk, node.ast,
        f'the target bitwidth {new_nbits} is less than the bitwidth of the operand ({old_nbits})!' )
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( new_nbits )
    node._is_explicit = True

  def visit_Truncate( s, node ):
    new_nbits = node.nbits
    child_type = node.value.Type
    old_nbits = child_type.get_dtype().get_length()
    if new_nbits > old_nbits:
      raise PyMTLTypeError( s.blk, node.ast,
        f'the target bitwidth {new_nbits} is larger than the bitwidth of the operand ({old_nbits})!' )
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( new_nbits )
    node._is_explicit = True

  def visit_Reduce( s, node ):
    child_type = node.value.Type
    node.Type = copy.copy( child_type )
    node.Type.dtype = rdt.Vector( 1 )
    node._is_explicit = True

  def visit_SizeCast( s, node ):
    nbits = node.nbits
    Type = node.value.Type

    # We do not check for bitwidth mismatch here because the user should
    # be able to explicitly convert signals/constatns to different bitwidth.
    node.Type = copy.copy( Type )
    node.Type.dtype = rdt.Vector( nbits )
    node._is_explicit = True

    try:
      node._value = node.value._value
    except AttributeError:
      pass

  def visit_Attribute( s, node ):
    # Attribute supported at L1: CurCompAttr
    if isinstance( node.value, bir.Base ):
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'type {node.value.Type} does not have attribute {node.attr}!' )
      s.accessed.add( node.attr )

    else:
      raise PyMTLTypeError( s.blk, node.ast,
        f'non-component attribute {node.attr} of type {node.value.Type} is not supported at L1!' )
    # value.attr has the type that is specified by the base
    node.Type = node.value.Type.get_property( node.attr )
    if isinstance( node.Type, rt.Const ):
      dtype = node.Type.get_dtype()
      if isinstance( dtype, rdt.Vector ):
        node._is_explicit = dtype.is_explicit()
      else:
        node._is_explicit = True
    else:
      node._is_explicit = True

  def _handle_index_extension( s, node, value, idx, dscp, inclusive=True ):
    expected_nbits = value.Type.get_index_width()
    idx_nbits = idx.Type.get_dtype().get_length()
    is_idx_reinterpretable = not idx._is_explicit

    if not inclusive and hasattr( idx, '_value' ):
      idx_nbits = s._get_nbits_from_value( idx._value-1 )

    if idx_nbits > expected_nbits:
      # Either bitwidth do not match or requires implicit truncation
      raise PyMTLTypeError( s.blk, node.ast,
        f'expects a {expected_nbits}-bit index but the given {dscp} has more ({idx_nbits}) bits!' )
    elif idx_nbits < expected_nbits:
      if is_idx_reinterpretable:
        # Implicit zero-extension
        s.enforcer.enter( s.blk, rt.NetWire(rdt.Vector(expected_nbits)), idx )
      else:
        # Bitwidth mismatch
        raise PyMTLTypeError( s.blk, node.ast,
          f'expects a {expected_nbits}-bit index but the given {dscp} has {idx_nbits} bits!' )
    else:
      if idx_nbits != idx.Type.get_dtype().get_length():
        # If we used a different bitwidth then enforce it
        s.enforcer.enter( s.blk, rt.NetWire(rdt.Vector(idx_nbits)), idx )

  def visit_Index( s, node ):
    idx = None if not hasattr(node.idx, "_value") else int(node.idx._value)
    if isinstance( node.value.Type, rt.Array ):
      if idx is not None and not (0 <= idx < node.value.Type.get_dim_sizes()[0]):
        raise PyMTLTypeError( s.blk, node.ast, 'array index out of range!' )
      node.Type = node.value.Type.get_next_dim_type()
      obj = node.value.Type.get_obj()
      s._handle_index_extension( node, node.value, node.idx, 'index' )

      # If the given index yields an integer, mark this node as implicit
      if idx is not None and obj is not None:
        if isinstance( node.Type, rt.Array ):
          node.Type.obj = obj[ int( idx ) ]
          node._is_explicit = True
        else:
          node._value = int( obj[ int( idx ) ] )
          node._is_explicit = False if isinstance(node._value, int) else True
      else:
        node._is_explicit = True

    elif isinstance( node.value.Type, rt.Signal ):
      dtype = node.value.Type.get_dtype()
      s._handle_index_extension( node, node.value, node.idx, 'index' )

      if node.value.Type.is_packed_indexable():
        if idx is not None and not (0 <= idx < dtype.get_length()):
          raise PyMTLTypeError( s.blk, node.ast,
            'bit selection index out of range!' )
        node.Type = node.value.Type.get_next_dim_type()
        node._is_explicit = True
      elif isinstance( dtype, rdt.Vector ):
        if idx is not None and not(0 <= idx < dtype.get_length()):
          raise PyMTLTypeError( s.blk, node.ast,
            'bit selection index out of range!' )
        node.Type = rt.NetWire( rdt.Vector( 1 ) )
        node._is_explicit = True
      else:
        raise PyMTLTypeError( s.blk, node.ast,
          f'cannot perform index on {dtype}!')

    else:
      # Should be unreachable
      raise PyMTLTypeError( s.blk, node.ast,
        f'cannot perform index on {node.value.Type}!')

  def visit_Slice( s, node ):
    lower_val = None if not hasattr(node.lower, "_value") else node.lower._value
    upper_val = None if not hasattr(node.upper, "_value") else node.upper._value
    dtype = node.value.Type.get_dtype()

    if hasattr(node.value, "_value"):
      raise PyMTLTypeError( s.blk, node.ast,
          f'cannot perform slicing on constant {node.value._value}!')

    if not isinstance( dtype, rdt.Vector ):
      raise PyMTLTypeError( s.blk, node.ast, f'cannot perform slicing on type {dtype}!')

    s._handle_index_extension( node, node.value, node.lower, 'slice lower bound' )
    s._handle_index_extension( node, node.value, node.upper, 'slice upper bound', False )

    if not lower_val is None and not upper_val is None:
      signal_nbits = dtype.get_length()
      # upper bound must be strictly larger than the lower bound
      if ( lower_val >= upper_val ):
        raise PyMTLTypeError( s.blk, node.ast,
          'the upper bound of a slice must be larger than the lower bound!' )
      # upper & lower bound should be less than the bit width of the signal
      if not ( 0 <= lower_val < upper_val <= signal_nbits ):
        raise PyMTLTypeError( s.blk, node.ast,
          'upper/lower bound of slice out of width of signal!' )
      node.Type = rt.NetWire( rdt.Vector( int( upper_val - lower_val ) ) )
      node._is_explicit = True

    else:
      # Try to special case the constant-stride part selection
      try:
        assert isinstance( node.upper, bir.BinOp )
        assert isinstance( node.upper.op, bir.Add )
        nbits = node.upper.right
        slice_size = nbits._value
        assert node.lower == node.upper.left
        node.Type = rt.NetWire( rdt.Vector( slice_size ) )
        node._is_explicit = True
        # Add new fields that might help translation
        node.size = slice_size
        node.base = node.lower
      except Exception:
        raise PyMTLTypeError( s.blk, node.ast, 'slice bounds must be constant!' )

  def _get_nbits_from_value( s, value ):
    if -1 <= value <= 1:
      return 1
    if value < 0:
      return math.ceil(math.log2(abs(value)))
    else:
      return math.ceil(math.log2(value+1))

#-------------------------------------------------------------------------
# Enforce types for all terms whose types are inferred (implicit)
#-------------------------------------------------------------------------

class BehavioralRTLIRTypeEnforcerL1( bir.BehavioralRTLIRNodeVisitor ):

  def __init__( s, component ):
    s.component = component

  def enter( s, blk, context, node ):
    s.blk = blk
    s.stack = deque([])
    with s.register_context( context ):
      s.visit( node )

  @contextmanager
  def register_context( s, context_type ):
    s.stack.append( context_type )
    yield
    s.stack.pop()

  def get_context( s, node, obj ):
    if not s.stack:
      raise PyMTLTypeError( s.blk, node.ast,
          f'no context was provided to validate the inferred bitwidth of {obj}!' )
    return s.stack[-1]

  def mutate_datatype( s, node, descp ):
    if not node._is_explicit:
      # assert isinstance(node.Type, rt.Const), f'internal error: {node} is not constant!'
      target_Type = s.get_context(node, descp).get_dtype()
      # All RTLIR datatypes are cached -- we don't want to invalidate the cache
      # and therefore a deepcopy is needed here
      node.Type = copy.deepcopy(node.Type)
      node.Type.dtype = target_Type
      # node._is_explicit = True

  def visit_FreeVar( s, node ):
    s.mutate_datatype( node, node.obj )

  def visit_Number( s, node ):
    s.mutate_datatype( node, node.value )

  def visit_Attribute( s, node ):
    s.mutate_datatype( node, node.attr )

  def visit_Index( s, node ):
    s.mutate_datatype( node, 'indexing' )



"""
========================================================================
HeuristicTopo.py
========================================================================

Author : Shunning Jiang
Date   : Feb 14, 2020
"""

import ast
from queue import PriorityQueue

from ..BasePass import BasePass, PassMetadata
from ..errors import PassOrderError
from ..sim.SimpleSchedulePass import SimpleSchedulePass, check_schedule
from .UnrollSimPass import UnrollSimPass

# FIXME also apply branchiness to all update_ff blocks

# Shunning feb-14-2020: This CountBranchesLoops is enhanced to recognize
# if a block only has loop at the top. So we found that a loop actually
# trace-breaks itselfs by calling call_assembler_r and return after the
# execution is complete. This means loops are totally different from
# branches.

class CountBranchesLoops( ast.NodeVisitor ):

  def enter( self, node ):
    self.num_br = 0
    self.loop_stack = 0
    self.only_loop_at_top = False
    self.visit( node )
    return self.num_br, self.only_loop_at_top

  def visit_FunctionDef( self, node ):
    for stmt in node.body:
      self.only_loop_at_top |= isinstance( stmt, (ast.For, ast.While) )

    for stmt in node.body:
      self.visit( stmt )

    if node.returns:
      for expr in node.returns:
        self.visit( expr )

  def visit_If( self, node ):
    self.only_loop_at_top &= (self.loop_stack > 0)

    # Special case "if s.reset:" -- it's only high for a cycle
    if isinstance( node.test, ast.Attribute ) and \
       node.test.attr == 'reset' and \
       isinstance( node.test.value, ast.Name ) and \
       node.test.value.id == 's':
      pass
    else:
      self.num_br += 1
    self.visit( node.test )

    for stmt in node.body:
      self.visit( stmt )

    for stmt in node.orelse:
      self.visit( stmt )

  def visit_IfExp( self, node ):
    self.only_loop_at_top &= (self.loop_stack > 0)

    # Special case "if s.reset:" -- it's only high for a cycle
    if isinstance( node.test, ast.Attribute ) and \
       node.test.attr == 'reset' and \
       isinstance( node.test.value, ast.Name ) and \
       node.test.value.id == 's':
      pass
    else:
      self.num_br += 1

    self.visit( node.test )
    self.visit( node.body )
    self.visit( node.orelse )

  # For/while is fine
  def visit_For( self, node ):
    self.loop_stack += 1
    # self.num_br += 0
    for stmt in node.body:
      self.visit( stmt )
    self.loop_stack -= 1

  def visit_While( self, node ):
    self.loop_stack += 1
    # self.num_br += 0
    for stmt in node.body:
      self.visit( stmt )
    self.loop_stack -= 1

class HeuristicTopoPass( UnrollSimPass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    if hasattr( top, "_sched" ):
      raise Exception("Some schedule pass has already been applied!")

    top._sched = PassMetadata()

    self.schedule_intra_cycle( top )

    # Reuse simple's ff and flip schedule
    simple = SimpleSchedulePass()
    simple.schedule_ff( top )
    simple.schedule_posedge_flip( top )

    top._sim = PassMetadata()

    self.create_print_line_trace( top )
    self.create_sim_cycle_count( top )
    self.create_lock_unlock_simulation( top )
    top.lock_in_simulation()

    self.create_sim_eval_comb( top )
    self.create_sim_tick( top )
    self.create_sim_reset( top )

  def schedule_intra_cycle( self, top ):

    # Construct the intra-cycle graph based on normal update blocks

    V   = top._dag.final_upblks - top.get_all_update_ff()
    E   = set()
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in top._dag.all_constraints: # u -> v
      if u in V and v in V:
        InD[v] += 1
        Es[u].append( v )
        E.add( (u, v) )

    # Extract branchiness

    # Initialize all generated net block to 0 branchiness
    branchiness = { x: 0 for x in top._dag.genblks }

    visitor = CountBranchesLoops()
    # FIXME use the pure-loop info
    for blk in top.get_all_update_blocks():
      hostobj = top.get_update_block_host_component( blk )
      branchiness[ blk ], _ = visitor.enter( hostobj.get_update_block_info( blk )[-1] )

    # Perform topological sort for a serial schedule.
    # Note that here we use a priority queue to get the blocks with small
    # branchiness as early as possible

    top._sched.update_schedule = update_schedule = []

    # Python3 doesn't have hash for functions
    id_v = { id(v): v for v in V}

    Q = PriorityQueue(0)
    for v in V:
      if not InD[v]:
        Q.put( (branchiness[ v ], id(v)) )

    while not Q.empty():
      br, u = Q.get()
      update_schedule.append( id_v[u] )
      for v in Es[id_v[u]]:
        InD[v] -= 1
        if not InD[v]:
          Q.put( (branchiness[ v ], id(v)) )

    check_schedule( top, update_schedule, V, E, InD )
